""" Implements class Structure to represent atomic structures.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""
import warnings
from copy import deepcopy
from collections import namedtuple
from typing import List, Dict, Tuple
from itertools import groupby
from dataclasses import dataclass
from os.path import splitext, basename

import numpy as np

from .parser import PDBParser, MMCIFParser
from .matching_utils import (
    rigid_transform,
    _format_mmcif_colunns,
    minimum_enclosing_box,
)
from .helpers import atom_profile
from .types import NDArray


@dataclass(repr=False)
class Structure:
    """Represents atomic structures in accordance with the Protein Data Bank (PDB)
    format specification.

    Attributes
    ----------
    record_type : NDArray
        Type of the record, e.g., ATOM, HETATM. Array shape = (n,)
    atom_serial_number : NDArray
        Serial number assigned to each atom.  Array shape = (n,)
    atom_name : NDArray
        Standardized names for each atom.  Array shape = (n,)
    atom_coordinate : NDArray
        The 3D Cartesian coordinates of each atom in x, y, z.  Array shape = (n,3 )
    alternate_location_indicator : NDArray
        Indicator for alternate locations of an atom if it exists in multiple places.
        Array shape = (n,)
    residue_name : NDArray
        Standard residue names where each atom belongs. Array shape = (n,)
    chain_identifier : NDArray
        Identifier for the chain where each atom is located. Array shape = (n,)
    residue_sequence_number : NDArray
        Sequence number of the residue in the protein chain for each atom.
        Array shape = (n,)
    code_for_residue_insertion : NDArray
        Code to denote any residue insertion. Array shape = (n,)
    occupancy : NDArray
        Occupancy factor of each atom, indicating the fraction of time the atom
        is located at its position. Array shape = (n,)
    temperature_factor : NDArray
        Measure of the atomic displacement or B-factor for each atom. Array shape = (n,)
    segment_identifier : NDArray
        Identifier for the segment where each atom belongs. Array shape = (n,)
    element_symbol : NDArray
        Atomic element symbol for each atom. Array shape = (n,)
    charge : NDArray
        Charge on the atom. Array shape = (n,)
    details : dict
        Any additional or auxiliary details. Array shape = (n,)

    References
    ----------
    .. [1]  https://www.cgl.ucsf.edu/chimera/docs/UsersGuide/tutorials/pdbintro.html
    """

    #: Return a numpy array with record types, e.g. ATOM, HETATM.
    record_type: NDArray

    #: Return a numpy array with serial number of each atom.
    atom_serial_number: NDArray

    #: Return a numpy array with name of each atom.
    atom_name: NDArray

    #: Return a numpy array with coordinates of each atom in x, y, z.
    atom_coordinate: NDArray

    #: Return a numpy array with alternate location indicates of each atom.
    alternate_location_indicator: NDArray

    #: Return a numpy array with originating residue names of each atom.
    residue_name: NDArray

    #: Return a numpy array with originating structure chain of each atom.
    chain_identifier: NDArray

    #: Return a numpy array with originating residue id of each atom.
    residue_sequence_number: NDArray

    #: Return a numpy array with insertion information d of each atom.
    code_for_residue_insertion: NDArray

    #: Return a numpy array with occupancy factors of each atom.
    occupancy: NDArray

    #: Return a numpy array with B-factors for each atom.
    temperature_factor: NDArray

    #: Return a numpy array with segment identifier for each atom.
    segment_identifier: NDArray

    #: Return a numpy array with element symbols of each atom.
    element_symbol: NDArray

    #: Return a numpy array with charges of each atom.
    charge: NDArray

    #: Returns a dictionary with class instance metadata.
    details: dict

    def __post_init__(self, *args, **kwargs):
        """
        Initialize the structure and populate header details.

        Raises
        ------
        ValueError
            If other NDArray attributes to not match the number of atoms.
            If the shape of atom_coordinates and chain_identifier doesn't match.
        """
        self._elements = Elements()
        self.details = self._populate_details(self.details)

        n_atoms = self.atom_coordinate.shape[0]
        for attribute in self.__dict__:
            value = getattr(self, attribute)
            if type(value) != np.ndarray:
                continue
            if value.shape[0] != n_atoms:
                raise ValueError(
                    f"Expected shape of {attribute}: {n_atoms}, got {value.shape[0]}."
                )

    def __getitem__(self, indices: List[int]) -> "Structure":
        """
        Get a Structure instance for specified indices.

        Parameters
        ----------
        indices : Union[int, bool, NDArray]
            The indices to get.

        Returns
        -------
        Structure
            The Structure instance for the given indices.
        """
        if type(indices) in (int, bool):
            indices = (indices,)

        indices = np.asarray(indices)
        attributes = (
            "record_type",
            "atom_serial_number",
            "atom_name",
            "atom_coordinate",
            "alternate_location_indicator",
            "residue_name",
            "chain_identifier",
            "residue_sequence_number",
            "code_for_residue_insertion",
            "occupancy",
            "temperature_factor",
            "segment_identifier",
            "element_symbol",
            "charge",
        )
        kwargs = {attr: getattr(self, attr)[indices] for attr in attributes}
        ret = self.__class__(**kwargs, details={})
        return ret

    def __repr__(self):
        """
        Return a string representation of the Structure.

        Returns
        -------
        str
            The string representation.
        """
        unique_chains = "-".join(
            [
                ",".join([str(x) for x in entity])
                for entity in self.details["unique_chains"]
            ]
        )
        min_atom = np.min(self.atom_serial_number)
        max_atom = np.max(self.atom_serial_number)
        n_atom = self.atom_serial_number.size

        min_residue = np.min(self.residue_sequence_number)
        max_residue = np.max(self.residue_sequence_number)
        n_residue = self.residue_sequence_number.size

        repr_str = (
            f"Structure object at {id(self)}\n"
            f"Unique Chains: {unique_chains}, "
            f"Atom Range: {min_atom}-{max_atom} [N = {n_atom}], "
            f"Residue Range: {min_residue}-{max_residue} [N = {n_residue}]"
        )
        return repr_str

    def get_chains(self) -> List[str]:
        """
        Returns a list of available chains.

        Returns
        -------
        list
            The list of available chains.
        """
        return list(self.details["chain_weight"].keys())

    def copy(self) -> "Structure":
        """
        Returns a copy of the Structure instance.

        Returns
        -------
        Structure
            The copied Structure instance.
        """
        return deepcopy(self)

    def _populate_details(self, details: Dict = {}) -> Dict:
        """
        Populate the details dictionary with the data from the Structure instance.

        Parameters
        ----------
        details : dict, optional
            The initial details dictionary, by default {}.

        Returns
        -------
        dict
            The populated details dictionary.
        """
        details["weight"] = np.sum(
            [self._elements[atype].atomic_weight for atype in self.element_symbol]
        )

        label, idx, chain = np.unique(
            self.chain_identifier, return_inverse=True, return_index=True
        )
        chain_weight = np.bincount(
            chain,
            [self._elements[atype].atomic_weight for atype in self.element_symbol],
        )
        labels = self.chain_identifier[idx]
        details["chain_weight"] = {key: val for key, val in zip(labels, chain_weight)}

        # Group non-unique chains in separate lists in details["unique_chains"]
        details["unique_chains"], temp = [], {}
        for chain_label in label:
            index = len(details["unique_chains"])
            chain_sequence = "".join(
                [
                    str(y)
                    for y in self.element_symbol[
                        np.where(self.chain_identifier == chain_label)
                    ]
                ]
            )
            if chain_sequence not in temp:
                temp[chain_sequence] = index
                details["unique_chains"].append([chain_label])
                continue
            idx = temp.get(chain_sequence)
            details["unique_chains"][idx].append(chain_label)

        filtered_data = [
            (label, integer)
            for label, integer in zip(
                self.chain_identifier, self.residue_sequence_number
            )
        ]
        filtered_data = sorted(filtered_data, key=lambda x: x[0])
        details["chain_range"] = {}
        for label, values in groupby(filtered_data, key=lambda x: x[0]):
            values = [int(x[1]) for x in values]
            details["chain_range"][label] = (min(values), max(values))

        return details

    @classmethod
    def from_file(
        cls,
        filename: str,
        keep_non_atom_records: bool = False,
        filter_by_elements: set = None,
        filter_by_residues: set = None,
    ) -> "Structure":
        """
        Reads in an mmcif or pdb file and converts it into class instance.

        Parameters
        ----------
        filename : str
            Path to the mmcif or pdb file.
        keep_non_atom_records : bool, optional
            Wheter to keep residues that are not labelled ATOM.
        filter_by_elements: set, optional
            Which elements to keep. Default corresponds to all elements.
        filter_by_residues: set, optional
            Which residues to keep. Default corresponds to all residues.

        Raises
        ------
        ValueError
            If the extension is not '.pdb' or '.cif'.

        Returns
        -------
        Structure
            Read in structure file.
        """
        _, file_extension = splitext(basename(filename.upper()))
        if file_extension == ".PDB":
            func = cls._load_pdb
        elif file_extension == ".CIF":
            func = cls._load_mmcif
        else:
            raise NotImplementedError(
                "Could not determine structure filetype from extension."
                " Supported filetypes are mmcif (.cif) and pdb (.pdb)."
            )
        data = func(filename)

        keep = np.ones(data["element_symbol"].size, dtype=bool)
        if filter_by_elements:
            keep = np.logical_and(
                keep,
                np.in1d(data["element_symbol"], np.array(list(filter_by_elements))),
            )
        if filter_by_residues:
            keep = np.logical_and(
                keep, np.in1d(data["residue_name"], np.array(list(filter_by_residues)))
            )
        if not keep_non_atom_records:
            keep = np.logical_and(keep, data["record_type"] == "ATOM")

        for key in data:
            if key == "details":
                continue
            if type(data[key]) == np.ndarray:
                data[key] = data[key][keep]
            else:
                data[key] = [x for x, flag in zip(data[key], keep) if flag]

        data["details"]["filepath"] = filename

        return cls(**data)

    @staticmethod
    def _load_mmcif(filename: str) -> Dict:
        """
        Parses a macromolecular Crystallographic Information File (mmCIF)
        and returns the data in a dictionary format.

        Parameters
        ----------
        filename : str
            The filename of the mmCIF to load.

        Returns
        -------
        dict
            A dictionary of numpy arrays. Keys are the names of the PDB
            coordinate section. In addition, some details about the parsed
            structure are included. In case of conversion failure, the failing
            attribute is set to 0 if its supposed to be an integer value.
        """
        result = MMCIFParser(filename)

        atom_site_mapping = {
            "record_type": ("group_PDB", str),
            "atom_serial_number": ("id", int),
            "atom_name": ("label_atom_id", str),
            "alternate_location_indicator": ("label_alt_id", str),
            "residue_name": ("label_comp_id", str),
            # "chain_identifier": ("auth_asym_id", str),
            "chain_identifier": ("label_asym_id", str),
            "residue_sequence_number": ("label_seq_id", int),
            "code_for_residue_insertion": ("pdbx_PDB_ins_code", str),
            "occupancy": ("occupancy", float),
            "temperature_factor": ("B_iso_or_equiv", float),
            "segment_identifier": ("pdbx_PDB_model_num", str),
            "element_symbol": ("type_symbol", str),
            "charge": ("pdbx_formal_charge", str),
        }

        out = {}
        for out_key, (atom_site_key, dtype) in atom_site_mapping.items():
            out_data = [
                x.strip() for x in result["atom_site"].get(atom_site_key, ["."])
            ]
            if dtype == int:
                out_data = [0 if x == "." else int(x) for x in out_data]
            try:
                out[out_key] = np.asarray(out_data).astype(dtype)
            except ValueError:
                default = ["."] if dtype == str else 0
                print(f"Converting {out_key} to {dtype} failed, set to {default}.")
                out[out_key] = np.repeat(default, len(out_data)).astype(dtype)

        number_entries = len(max(out.values(), key=len))
        for key, value in out.items():
            if value.size != 1:
                continue
            out[key] = np.repeat(value, number_entries // value.size)

        out["details"] = {}
        out["atom_coordinate"] = np.transpose(
            np.array(
                [
                    result["atom_site"]["Cartn_x"],
                    result["atom_site"]["Cartn_y"],
                    result["atom_site"]["Cartn_z"],
                ],
                dtype=np.float32,
            )
        )

        detail_mapping = {
            "resolution": ("em_3d_reconstruction", "resolution", np.nan),
            "resolution_method": ("em_3d_reconstruction", "resolution_method", np.nan),
            "method": ("exptl", "method", np.nan),
            "electron_source": ("em_imaging", "electron_source", np.nan),
            "illumination_mode": ("em_imaging", "illumination_mode", np.nan),
            "microscope_model": ("em_imaging", "microscope_model", np.nan),
        }
        for out_key, (base_key, inner_key, default) in detail_mapping.items():
            if base_key not in result:
                continue
            out["details"][out_key] = result[base_key].get(inner_key, default)

        return out

    @staticmethod
    def _load_pdb(filename: str) -> Dict:
        """
        Parses a Protein Data Bank (PDB) file and returns the data
        in a dictionary format.

        Parameters
        ----------
        filename : str
            The filename of the PDB file to load.

        Returns
        -------
        dict
            A dictionary of numpy arrays. Keys are the names of the PDB
            coordinate section. In addition, some details about the parsed
            structure are included. In case of conversion failure, the failing
            attribute is set to 0 if its supposed to be an integer value.
        """
        result = PDBParser(filename)

        atom_site_mapping = {
            "record_type": ("record_type", str),
            "atom_serial_number": ("atom_serial_number", int),
            "atom_name": ("atom_name", str),
            "alternate_location_indicator": ("alternate_location_indicator", str),
            "residue_name": ("residue_name", str),
            "chain_identifier": ("chain_identifier", str),
            "residue_sequence_number": ("residue_sequence_number", int),
            "code_for_residue_insertion": ("code_for_residue_insertion", str),
            "occupancy": ("occupancy", float),
            "temperature_factor": ("temperature_factor", float),
            "segment_identifier": ("segment_identifier", str),
            "element_symbol": ("element_symbol", str),
            "charge": ("charge", str),
        }

        out = {"details": result["details"]}
        for out_key, (inner_key, dtype) in atom_site_mapping.items():
            out_data = [x.strip() for x in result[inner_key]]
            if dtype == int:
                out_data = [0 if x == "." else int(x) for x in out_data]
            try:
                out[out_key] = np.asarray(out_data).astype(dtype)
            except ValueError:
                default = "." if dtype == str else 0
                print(
                    f"Converting {out_key} to {dtype} failed. Setting {out_key} to {default}."
                )
                out[out_key] = np.repeat(default, len(out_data)).astype(dtype)

        out["atom_coordinate"] = np.array(result["atom_coordinate"], dtype=np.float32)

        return out

    def to_file(self, filename: str) -> None:
        """
        Writes the Structure instance data to a Protein Data Bank (PDB) or
        macromolecular Crystallographic Information File (mmCIF) file depending
        one whether filename ends with '.pdb' or '.cif'.

        Raises
        ------
        ValueError
            If the extension is not '.pdb' or '.cif'.

        Parameters
        ----------
        filename : str
            The filename of the file to write.
        """
        data_out = []
        if np.any(np.vectorize(len)(self.chain_identifier) > 2):
            warnings.warn("Chain identifiers longer than one will be shortened.")

        _, file_extension = splitext(basename(filename.upper()))
        if file_extension == ".PDB":
            func = self._write_pdb
        elif file_extension == ".CIF":
            func = self._write_mmcif
        else:
            raise NotImplementedError(
                "Could not determine structure filetype."
                " Supported filetypes are mmcif (.cif) and pdb (.pdb)."
            )

        if self.atom_coordinate.shape[0] > 10**5 and func == self._write_pdb:
            warnings.warn(
                "The structure contains more than 100,000 atoms. Consider using mmcif."
            )

        with open(filename, mode="w", encoding="utf-8") as ofile:
            ofile.writelines(func())

    def _write_pdb(self) -> List[str]:
        """
        Returns a PDB string representation of the structure instance.

        Returns
        -------
        list
            List containing PDB file coordine lines.
        """
        data_out = []
        for index in range(self.atom_coordinate.shape[0]):
            x, y, z = self.atom_coordinate[index, :]
            line = list(" " * 80)
            line[0:6] = f"{self.record_type[index]:<6}"
            line[6:11] = f"{self.atom_serial_number[index]:>5}"
            line[12:16] = f"{self.atom_name[index]:<4}"
            line[16] = f"{self.alternate_location_indicator[index]:<1}"
            line[17:20] = f"{self.residue_name[index]:<3}"
            line[21] = f"{self.chain_identifier[index][0]:<1}"
            line[22:26] = f"{self.residue_sequence_number[index]:>4}"
            line[26] = f"{self.code_for_residue_insertion[index]:<1}"
            line[30:38] = f"{x:>8.3f}"
            line[38:46] = f"{y:>8.3f}"
            line[46:54] = f"{z:>8.3f}"
            line[54:60] = f"{self.occupancy[index]:>6.2f}"
            line[60:66] = f"{self.temperature_factor[index]:>6.2f}"
            line[72:76] = f"{self.segment_identifier[index]:>4}"
            line[76:78] = f"{self.element_symbol[index]:<2}"
            line[78:80] = f"{self.charge[index]:>2}"
            data_out.append("".join(line))
        data_out.append("END")
        data_out = "\n".join(data_out)
        return data_out

    def _write_mmcif(self) -> List[str]:
        """
        Returns a MMCIF string representation of the structure instance.

        Returns
        -------
        list
            List containing MMCIF file coordinate lines.
        """
        model_num, entity_id = 1, 1
        data = {
            "group_PDB": [],
            "id": [],
            "type_symbol": [],
            "label_atom_id": [],
            "label_alt_id": [],
            "label_comp_id": [],
            "label_asym_id": [],
            "label_entity_id": [],
            "label_seq_id": [],
            "pdbx_PDB_ins_code": [],
            "Cartn_x": [],
            "Cartn_y": [],
            "Cartn_z": [],
            "occupancy": [],
            "B_iso_or_equiv": [],
            "pdbx_formal_charge": [],
            "auth_seq_id": [],
            "auth_comp_id": [],
            "auth_asym_id": [],
            "auth_atom_id": [],
            "pdbx_PDB_model_num": [],
        }

        for index in range(self.atom_coordinate.shape[0]):
            x, y, z = self.atom_coordinate[index, :]
            data["group_PDB"].append(self.record_type[index])
            data["id"].append(str(self.atom_serial_number[index]))
            data["type_symbol"].append(self.element_symbol[index])
            data["label_atom_id"].append(self.atom_name[index])
            data["label_alt_id"].append(self.alternate_location_indicator[index])
            data["label_comp_id"].append(self.residue_name[index])
            data["label_asym_id"].append(self.chain_identifier[index][0])
            data["label_entity_id"].append(str(entity_id))
            data["label_seq_id"].append(str(self.residue_sequence_number[index]))
            data["pdbx_PDB_ins_code"].append(self.code_for_residue_insertion[index])
            data["Cartn_x"].append(f"{x:.3f}")
            data["Cartn_y"].append(f"{y:.3f}")
            data["Cartn_z"].append(f"{z:.3f}")
            data["occupancy"].append(f"{self.occupancy[index]:.2f}")
            data["B_iso_or_equiv"].append(f"{self.temperature_factor[index]:.2f}")
            data["pdbx_formal_charge"].append(self.charge[index])
            data["auth_seq_id"].append(str(self.residue_sequence_number[index]))
            data["auth_comp_id"].append(self.residue_name[index])
            data["auth_asym_id"].append(self.chain_identifier[index][0])
            data["auth_atom_id"].append(self.atom_name[index])
            data["pdbx_PDB_model_num"].append(str(model_num))

        output_data = {"atom_site": data}
        original_file = self.details.get("filepath", "")
        try:
            new_data = {k: v for k, v in MMCIFParser(original_file).items()}
            index = self.atom_serial_number - 1
            new_data["atom_site"] = {
                k: [v[i] for i in index] for k, v in new_data["atom_site"].items()
            }
            new_data["atom_site"]["Cartn_x"] = data["Cartn_x"]
            new_data["atom_site"]["Cartn_y"] = data["Cartn_y"]
            new_data["atom_site"]["Cartn_z"] = data["Cartn_z"]
            output_data = new_data
        except Exception:
            pass

        ret = ""
        for category, subdict in output_data.items():
            ret += "#\n"
            is_loop = isinstance(subdict[list(subdict.keys())[0]], list)
            if not is_loop:
                for k in subdict:
                    ret += f"_{category}.{k}\t{subdict[k]}\n"
            else:
                ret += "loop_\n"
                ret += "".join([f"_{category}.{k}\n" for k in subdict])
                padded_subdict = _format_mmcif_colunns(subdict)

                data = [
                    "".join([str(x) for x in content])
                    for content in zip(*padded_subdict.values())
                ]
                ret += "\n".join([entry for entry in data]) + "\n"

        return ret

    def subset_by_chain(self, chain: str = None) -> "Structure":
        """
        Return a subset of the structure that contains only atoms belonging to
        a specific chain. If no chain is specified, all chains are returned.

        Parameters
        ----------
        chain : str, optional
            The chain identifier. If multiple chains should be selected they need
            to be a comma separated string, e.g. 'A,B,CE'. If chain None,
            all chains are returned. Default is None.

        Returns
        -------
        Structure
            A subset of the original structure containing only the specified chain.
        """
        chain = np.unique(self.chain_identifier) if chain is None else chain.split(",")
        keep = np.in1d(self.chain_identifier, chain)
        return self[keep]

    def subset_by_range(
        self,
        start: int,
        stop: int,
        chain: str = None,
    ) -> "Structure":
        """
        Return a subset of the structure within a specific range of residues.

        Parameters
        ----------
        start : int
            The starting residue sequence number.

        stop : int
            The ending residue sequence number.

        chain : str, optional
            The chain identifier. If multiple chains should be selected they need
            to be a comma separated string, e.g. 'A,B,CE'. If chain None,
            all chains are returned. Default is None.

        Returns
        -------
        Structure
            A subset of the original structure within the specified residue range.
        """
        ret = self.subset_by_chain(chain=chain)
        keep = np.logical_and(
            ret.residue_sequence_number >= start, ret.residue_sequence_number <= stop
        )
        return ret[keep]

    def center_of_mass(self) -> NDArray:
        """
        Calculate the center of mass of the structure.

        Returns
        -------
        NDArray
            The center of mass of the structure.
        """
        weights = [self._elements[atype].atomic_weight for atype in self.element_symbol]
        return np.dot(self.atom_coordinate.T, weights) / np.sum(weights)

    def rigid_transform(
        self,
        rotation_matrix: NDArray,
        translation: NDArray,
        use_geometric_center: bool = False,
    ) -> "Structure":
        """
        Performs a rigid transform of internal structure coordinates.

        Parameters
        ----------
        rotation_matrix : NDArray
            The rotation matrix to apply to the coordinates.
        translation : NDArray
            The vector to translate the coordinates by.
        use_geometric_center : bool, optional
            Whether to use geometric or coordinate center.

        Returns
        -------
        Structure
            The transformed instance of :py:class:`tme.structure.Structure`.
        """
        out = np.empty_like(self.atom_coordinate.T)
        rigid_transform(
            coordinates=self.atom_coordinate.T,
            rotation_matrix=rotation_matrix,
            translation=translation,
            out=out,
            use_geometric_center=use_geometric_center,
        )
        ret = self.copy()
        ret.atom_coordinate = out.T.copy()
        return ret

    def centered(self) -> Tuple["Structure", NDArray]:
        """
        Shifts the structure analogous to :py:meth:`tme.density.Density.centered`.

        Returns
        -------
        Structure
            A copy of the class instance whose data center of mass is in the
            center of the data array.
        NDArray
            The coordinate translation.

        See Also
        --------
        :py:meth:`tme.Density.centered`
        """
        center_of_mass = self.center_of_mass()
        enclosing_box = minimum_enclosing_box(coordinates=self.atom_coordinate.T)
        shift = np.subtract(np.divide(enclosing_box, 2), center_of_mass)

        transformed_structure = self.rigid_transform(
            translation=shift, rotation_matrix=np.eye(shift.size)
        )

        return transformed_structure, shift

    def _coordinate_to_position(
        self,
        shape: Tuple[int],
        sampling_rate: Tuple[float],
        origin: Tuple[float],
    ) -> (NDArray, Tuple[str], Tuple[int], float, Tuple[float]):
        """
        Converts coordinates to positions.

        Parameters
        ----------
        shape : Tuple[int,]
            The desired shape of the output array.

        sampling_rate : float
            The sampling rate of the output array in unit of self.atom_coordinate.

        origin : Tuple[float,]
            The origin of the coordinate system.
        Returns
        -------
        Tuple[NDArray, List[str], Tuple[int, ], float, Tuple[float,]]
            Returns positions, atom_types, shape, sampling_rate, and origin.
        """
        coordinates = self.atom_coordinate.copy()
        atom_types = self.element_symbol.copy()

        # positions are in x, y, z map is z, y, x
        coordinates = coordinates[:, ::-1]

        sampling_rate = 1 if sampling_rate is None else sampling_rate
        adjust_origin = origin is not None and shape is None
        origin = coordinates.min(axis=0) if origin is None else origin
        positions = (coordinates - origin) / sampling_rate
        positions = np.rint(positions).astype(int)

        if adjust_origin:
            left_shift = positions.min(axis=0)
            positions -= left_shift
            shape = positions.max(axis=0) + 1
            origin = origin + np.multiply(left_shift, sampling_rate)

        if shape is None:
            shape = positions.max(axis=0) + 1

        valid_positions = np.sum(
            np.logical_and(positions < shape, positions >= 0), axis=1
        )

        positions = positions[valid_positions == positions.shape[1], :]
        atom_types = atom_types[valid_positions == positions.shape[1]]

        self.details["nAtoms_outOfBound"] = 0
        if positions.shape[0] != coordinates.shape[0]:
            out_of_bounds = coordinates.shape[0] - positions.shape[0]
            print(f"{out_of_bounds}/{coordinates.shape[0]} atoms were out of bounds.")
            self.details["nAtoms_outOfBound"] = out_of_bounds

        return positions, atom_types, shape, sampling_rate, origin

    def _position_to_vdw_sphere(
        self,
        positions: Tuple[float],
        atoms: Tuple[str],
        sampling_rate: Tuple[float],
        volume: NDArray,
    ) -> None:
        """
        Updates a volume with van der Waals spheres.

        Parameters
        ----------
        positions : Tuple[float, float, float]
            The positions of the atoms.

        atoms : Tuple[str]
            The types of the atoms.

        sampling_rate : float
            The desired sampling rate in unit of self.atom_coordinate of the
            output array.

        volume : NDArray
            The volume to update.
        """
        index_dict, vdw_rad, shape = {}, {}, volume.shape
        for atom_index, atom_position in enumerate(positions):
            atom_type = atoms[atom_index]
            if atom_type not in index_dict.keys():
                atom_vdwr = np.ceil(
                    np.divide(self._elements[atom_type].vdwr, (sampling_rate * 100))
                ).astype(int)

                vdw_rad[atom_type] = atom_vdwr
                atom_slice = tuple(slice(-k, k + 1) for k in atom_vdwr)
                distances = np.linalg.norm(
                    np.divide(
                        np.mgrid[atom_slice],
                        atom_vdwr.reshape((-1,) + (1,) * volume.ndim),
                    ),
                    axis=0,
                )
                index_dict[atom_type] = (distances <= 1).astype(volume.dtype)

            footprint = index_dict[atom_type]
            start = np.maximum(np.subtract(atom_position, vdw_rad[atom_type]), 0)
            stop = np.minimum(np.add(atom_position, vdw_rad[atom_type]) + 1, shape)
            volume_slice = tuple(slice(*coord) for coord in zip(start, stop))

            start_index = np.maximum(-np.subtract(atom_position, vdw_rad[atom_type]), 0)
            stop_index = np.add(
                footprint.shape,
                np.minimum(
                    np.subtract(shape, np.add(atom_position, vdw_rad[atom_type]) + 1), 0
                ),
            )
            index_slice = tuple(slice(*coord) for coord in zip(start_index, stop_index))
            volume[volume_slice] += footprint[index_slice]

    def _position_to_scattering_factors(
        self,
        positions: NDArray,
        atoms: NDArray,
        sampling_rate: NDArray,
        volume: NDArray,
        lowpass_filter: bool = True,
        downsampling_factor: float = 1.35,
        source: str = "peng1995",
    ) -> None:
        """
        Updates a volume with scattering factors.

        Parameters
        ----------
        positions : NDArray
            The positions of the atoms.
        atoms : NDArray
            Element symbols.
        sampling_rate : float
            Sampling rate that was used to convert coordinates to positions.
        volume : NDArray
            The volume to update.
        lowpass_filter : NDArray
            Whether the scattering factors hsould be lowpass filtered.
        downsampling_factor : NDArray
            Downsampling factor for scattering factor computation.
        source : str
            Which scattering factors to use

        Reference
        ---------
        https://github.com/I2PC/xmipp.
        """
        scattering_profiles, shape = dict(), volume.shape
        for atom_index, point in enumerate(positions):
            if atoms[atom_index] not in scattering_profiles:
                spline = atom_profile(
                    atom=atoms[atom_index],
                    M=downsampling_factor,
                    method=source,
                    lfilter=lowpass_filter,
                )
                scattering_profiles.update({atoms[atom_index]: spline})

            atomic_radius = np.divide(
                self._elements[atoms[atom_index]].vdwr, sampling_rate * 100
            )
            starts = np.maximum(np.ceil(point - atomic_radius), 0).astype(int)
            stops = np.minimum(np.floor(point + atomic_radius), shape).astype(int)

            grid_index = np.meshgrid(
                *[range(start, stop) for start, stop in zip(starts, stops)]
            )
            distances = np.einsum(
                "aijk->ijk",
                np.array([(grid_index[i] - point[i]) ** 2 for i in range(len(point))]),
                dtype=np.float64,
            )
            distances = np.sqrt(distances)
            if not len(distances):
                grid_index, distances = point, 0
            np.add.at(
                volume,
                tuple(grid_index),
                scattering_profiles[atoms[atom_index]](distances),
            )

    def _get_atom_weights(
        self, atoms: Tuple[str] = None, weight_type: str = "atomic_weight"
    ) -> Tuple[float]:
        """
        Returns weights of individual atoms according to a specified weight type.

        Parameters
        ----------
        atoms : Tuple of strings, optional
            The atoms to get the weights for. If None, weights for all atoms
            are used. Default is None.

        weight_type : str, optional
            The type of weights to return. This can either be 'atomic_weight',
            'atomic_number', or 'van_der_waals_radius'. Default is 'atomic_weight'.

        Returns
        -------
        List[float]
            A list containing the weights of the atoms.
        """
        atoms = self.element_symbol if atoms is None else atoms
        match weight_type:
            case "atomic_weight":
                weight = [self._elements[atom].atomic_weight for atom in atoms]
            case "atomic_number":
                weight = [self._elements[atom].atomic_number for atom in atoms]
            case _:
                raise NotImplementedError(
                    "weight_type can either be 'atomic_weight' or 'atomic_number'"
                )
        return weight

    def to_volume(
        self,
        shape: Tuple[int] = None,
        sampling_rate: NDArray = None,
        origin: Tuple[float] = None,
        chain: str = None,
        weight_type: str = "atomic_weight",
        scattering_args: Dict = dict(),
    ) -> Tuple[NDArray, Tuple[int], NDArray]:
        """
        Converts atom coordinates of shape [n x 3] x, y, z to a volume with
        index z, y, x.

        Parameters
        ----------
        shape : Tuple[int, ...], optional
            Desired shape of the output array. If shape is given its expected to be
            in z, y, x form.
        sampling_rate : float, optional
            Sampling rate of the output array in the unit of self.atom_coordinate
        origin : Tuple[float, ...], optional
            Origin of the coordinate system. If origin is given its expected to be
            in z, y, x form.
        chain : str, optional
            The chain identifier. If multiple chains should be selected they need
            to be a comma separated string, e.g. 'A,B,CE'. If chain None,
            all chains are returned. Default is None.
        weight_type : str, optional
            Which weight should be given to individual atoms.
        scattering_args : dict, optional
            Additional arguments for scattering factor computation.

        Returns
        -------
        Tuple[NDArray, Tuple[int], NDArray]
            The volume, its origin and the voxel size in Ångstrom.
        """
        _weight_types = {
            "atomic_weight",
            "atomic_number",
            "van_der_waals_radius",
            "scattering_factors",
            "lowpass_scattering_factors",
        }
        _weight_string = ",".join([f"'{x}'" for x in _weight_types])
        if weight_type not in _weight_types:
            raise NotImplementedError(f"weight_type needs to be in {_weight_string}")

        if sampling_rate is None:
            sampling_rate = np.ones(self.atom_coordinate.shape[1])
        sampling_rate = np.array(sampling_rate)
        if sampling_rate.size == 1:
            sampling_rate = np.repeat(sampling_rate, self.atom_coordinate.shape[1])
        elif sampling_rate.size != self.atom_coordinate.shape[1]:
            raise ValueError(
                "sampling_rate should either be single value of array with"
                f"size {self.atom_coordinate.shape[1]}."
            )
        if "source" not in scattering_args:
            scattering_args["source"] = "peng1995"

        temp = self.subset_by_chain(chain=chain)

        positions, atoms, shape, sampling_rate, origin = temp._coordinate_to_position(
            shape=shape, sampling_rate=sampling_rate, origin=origin
        )
        volume = np.zeros(shape, dtype=np.float32)
        if weight_type in ("atomic_weight", "atomic_number"):
            weights = temp._get_atom_weights(atoms=atoms, weight_type=weight_type)
            np.add.at(volume, tuple(positions.T), weights)
        elif weight_type == "van_der_waals_radius":
            self._position_to_vdw_sphere(positions, atoms, sampling_rate, volume)
        elif weight_type == "scattering_factors":
            self._position_to_scattering_factors(
                positions,
                atoms,
                sampling_rate,
                volume,
                lowpass_filter=False,
                **scattering_args,
            )
        elif weight_type == "lowpass_scattering_factors":
            self._position_to_scattering_factors(
                positions,
                atoms,
                sampling_rate,
                volume,
                lowpass_filter=True,
                **scattering_args,
            )

        self.details.update(temp.details)
        return volume, origin, sampling_rate

    @classmethod
    def compare_structures(
        cls,
        structure1: "Structure",
        structure2: "Structure",
        origin: NDArray = None,
        sampling_rate: float = None,
        weighted: bool = False,
    ) -> float:
        """
        Compute root mean square deviation (RMSD) between two structures.

        Both structures need to have the same number of atoms. In practice, this means
        that *structure2* is a transformed version of *structure1*

        Parameters
        ----------
        structure1 : Structure
            Structure 1.

        structure2 : Structure
            Structure 2.

        origin : NDArray, optional
            Origin of the structure coordinate system.

        sampling_rate : float, optional
            Sampling rate if discretized on a grid in the unit of self.atom_coordinate.

        weighted : bool, optional
            Whether atoms should be weighted by their atomic weight.

        Returns
        -------
        float
            Root Mean Square Deviation (RMSD)
        """
        if origin is None:
            origin = np.zeros(structure1.atom_coordinate.shape[1])

        coordinates1 = structure1.atom_coordinate
        coordinates2 = structure2.atom_coordinate
        atoms1, atoms2 = structure1.element_symbol, structure2.element_symbol
        if sampling_rate is not None:
            coordinates1 = np.rint((coordinates1 - origin) / sampling_rate).astype(int)
            coordinates2 = np.rint((coordinates2 - origin) / sampling_rate).astype(int)

        weights1 = np.array(structure1._get_atom_weights(atoms=atoms1))
        weights2 = np.array(structure2._get_atom_weights(atoms=atoms2))
        if not weighted:
            weights1 = np.ones_like(weights1)
            weights2 = np.ones_like(weights2)

        if not np.allclose(coordinates1.shape, coordinates2.shape):
            raise ValueError(
                "Input structures need to have the same number of coordinates."
            )
        if not np.allclose(weights1.shape, weights2.shape):
            raise ValueError("Input structures need to have the same number of atoms.")

        squared_diff = np.sum(np.square(coordinates1 - coordinates2), axis=1)
        weighted_quared_diff = squared_diff * ((weights1 + weights2) / 2)
        rmsd = np.sqrt(np.mean(weighted_quared_diff))

        return rmsd

    @classmethod
    def align_structures(
        cls,
        structure1: "Structure",
        structure2: "Structure",
        origin: NDArray = None,
        sampling_rate: float = None,
        weighted: bool = False,
    ) -> Tuple["Structure", float]:
        """
        Align the atom coordinates of structure2 to structure1 using
        the Kabsch algorithm.

        Both structures need to have the same number of atoms. In practice, this means
        that *structure2* is a subset of *structure1*

        Parameters
        ----------
        structure1 : Structure
            Structure 1.

        structure2 : Structure
            Structure 2.

        origin : NDArray, optional
            Origin of the structure coordinate system.

        sampling_rate : float, optional
            Voxel size if discretized on a grid.

        weighted : bool, optional
            Whether atoms should be weighted by their atomic weight.

        Returns
        -------
        Structure
            *structure2* aligned to *structure1*.
        float
            Root Mean Square Error (RMSE)
        """
        if origin is None:
            origin = np.minimum(
                structure1.atom_coordinate.min(axis=0),
                structure2.atom_coordinate.min(axis=0),
            ).astype(int)

        initial_rmsd = cls.compare_structures(
            structure1=structure1,
            structure2=structure2,
            origin=origin,
            sampling_rate=sampling_rate,
            weighted=weighted,
        )

        reference = structure1.atom_coordinate.copy()
        query = structure2.atom_coordinate.copy()
        if sampling_rate is not None:
            reference, atoms1, shape, _, _ = structure1._coordinate_to_position(
                shape=None, sampling_rate=sampling_rate, origin=origin
            )
            query, atoms2, shape, _, _ = structure2._coordinate_to_position(
                shape=None, sampling_rate=sampling_rate, origin=origin
            )

        reference_mean = reference.mean(axis=0)
        query_mean = query.mean(axis=0)

        reference = reference - reference_mean
        query = query - query_mean

        corr = np.dot(query.T, reference)
        U, S, Vh = np.linalg.svd(corr)

        rotation = np.dot(Vh.T, U.T).T
        if np.linalg.det(rotation) < 0:
            Vh[2, :] *= -1
            rotation = np.dot(Vh.T, U.T).T

        translation = reference_mean - np.dot(query_mean, rotation)

        temp = structure1.copy()
        temp.atom_coordinate = reference + reference_mean
        ret = structure2.copy()
        ret.atom_coordinate = np.dot(query + query_mean, rotation) + translation

        final_rmsd = cls.compare_structures(
            structure1=temp,
            structure2=ret,
            origin=origin,
            sampling_rate=None,
            weighted=weighted,
        )

        print(f"Initial RMSD: {initial_rmsd:.5f} - Final RMSD: {final_rmsd:.5f}")

        return ret, final_rmsd


@dataclass(frozen=True, repr=True)
class Elements:
    """
    Lookup table containing information on chemical elements.
    """

    Atom = namedtuple(
        "Atom",
        [
            "atomic_number",
            "atomic_radius",
            "lattice_constant",
            "lattice_structure",
            "vdwr",
            "covalent_radius_bragg",
            "atomic_weight",
        ],
    )
    _default = Atom(0, 0, 0, "Atom does not exist in ressource.", 0, 0, 0)
    _elements = {
        "H": Atom(1, 25, 3.75, "HEX", 110, np.nan, 1.008),
        "HE": Atom(2, 120, 3.57, "HEX", 140, np.nan, 4.002602),
        "LI": Atom(3, 145, 3.49, "BCC", 182, 150, 6.94),
        "BE": Atom(4, 105, 2.29, "HEX", 153, 115, 9.0121831),
        "B": Atom(5, 85, 8.73, "TET", 192, np.nan, 10.81),
        "C": Atom(6, 70, 3.57, "DIA", 170, 77, 12.011),
        "N": Atom(7, 65, 4.039, "HEX", 155, 65, 14.007),
        "O": Atom(8, 60, 6.83, "CUB", 152, 65, 15.999),
        "F": Atom(9, 50, np.nan, "MCL", 147, 67, 18.998403163),
        "NE": Atom(10, 160, 4.43, "FCC", 154, np.nan, 20.1797),
        "NA": Atom(11, 180, 4.23, "BCC", 227, 177, 22.98976928),
        "MG": Atom(12, 150, 3.21, "HEX", 173, 142, 24.305),
        "AL": Atom(13, 125, 4.05, "FCC", 184, 135, 26.9815385),
        "SI": Atom(14, 110, 5.43, "DIA", 210, 117, 28.085),
        "P": Atom(15, 100, 7.17, "CUB", 180, np.nan, 30.973761998),
        "S": Atom(16, 100, 10.47, "ORC", 180, 102, 32.06),
        "CL": Atom(17, 100, 6.24, "ORC", 175, 105, 35.45),
        "AR": Atom(18, 71, 5.26, "FCC", 188, np.nan, 39.948),
        "K": Atom(19, 220, 5.23, "BCC", 275, 207, 39.0983),
        "CA": Atom(20, 180, 5.58, "FCC", 231, 170, 40.078),
        "SC": Atom(21, 160, 3.31, "HEX", 215, np.nan, 44.955908),
        "TI": Atom(22, 140, 2.95, "HEX", 211, 140, 47.867),
        "V": Atom(23, 135, 3.02, "BCC", 207, np.nan, 50.9415),
        "CR": Atom(24, 140, 2.88, "BCC", 206, 140, 51.9961),
        "MN": Atom(25, 140, 8.89, "CUB", 205, 147, 54.938044),
        "FE": Atom(26, 140, 2.87, "BCC", 204, 140, 55.845),
        "CO": Atom(27, 135, 2.51, "HEX", 200, 137, 58.933194),
        "NI": Atom(28, 135, 3.52, "FCC", 197, 135, 58.6934),
        "CU": Atom(29, 135, 3.61, "FCC", 196, 137, 63.546),
        "ZN": Atom(30, 135, 2.66, "HEX", 201, 132, 65.38),
        "GA": Atom(31, 130, 4.51, "ORC", 187, np.nan, 69.723),
        "GE": Atom(32, 125, 5.66, "DIA", 211, np.nan, 72.63),
        "AS": Atom(33, 115, 4.13, "RHL", 185, 126, 74.921595),
        "SE": Atom(34, 115, 4.36, "HEX", 190, 117, 78.971),
        "BR": Atom(35, 115, 6.67, "ORC", 185, 119, 79.904),
        "KR": Atom(36, np.nan, 5.72, "FCC", 202, np.nan, 83.798),
        "RB": Atom(37, 235, 5.59, "BCC", 303, 225, 85.4678),
        "SR": Atom(38, 200, 6.08, "FCC", 249, 195, 87.62),
        "Y": Atom(39, 180, 3.65, "HEX", 232, np.nan, 88.90584),
        "ZR": Atom(40, 155, 3.23, "HEX", 223, np.nan, 91.224),
        "NB": Atom(41, 145, 3.3, "BCC", 218, np.nan, 92.90637),
        "MO": Atom(42, 145, 3.15, "BCC", 217, np.nan, 95.95),
        "TC": Atom(43, 135, 2.74, "HEX", 216, np.nan, 97.90721),
        "RU": Atom(44, 130, 2.7, "HEX", 213, np.nan, 101.07),
        "RH": Atom(45, 135, 3.8, "FCC", 210, np.nan, 102.9055),
        "PD": Atom(46, 140, 3.89, "FCC", 210, np.nan, 106.42),
        "AG": Atom(47, 160, 4.09, "FCC", 211, 177, 107.8682),
        "CD": Atom(48, 155, 2.98, "HEX", 218, 160, 112.414),
        "IN": Atom(49, 155, 4.59, "TET", 193, np.nan, 114.818),
        "SN": Atom(50, 145, 5.82, "TET", 217, 140, 118.71),
        "SB": Atom(51, 145, 4.51, "RHL", 206, 140, 121.76),
        "TE": Atom(52, 140, 4.45, "HEX", 206, 133, 127.6),
        "I": Atom(53, 140, 7.72, "ORC", 198, 140, 126.90447),
        "XE": Atom(54, np.nan, 6.2, "FCC", 216, np.nan, 131.293),
        "CS": Atom(55, 260, 6.05, "BCC", 343, 237, 132.90545196),
        "BA": Atom(56, 215, 5.02, "BCC", 268, 210, 137.327),
        "LA": Atom(57, 195, 3.75, "HEX", 243, np.nan, 138.90547),
        "CE": Atom(58, 185, 5.16, "FCC", 242, np.nan, 140.116),
        "PR": Atom(59, 185, 3.67, "HEX", 240, np.nan, 140.90766),
        "ND": Atom(60, 185, 3.66, "HEX", 239, np.nan, 144.242),
        "PM": Atom(61, 185, np.nan, "", 238, np.nan, 144.91276),
        "SM": Atom(62, 185, 9, "RHL", 236, np.nan, 150.36),
        "EU": Atom(63, 185, 4.61, "BCC", 235, np.nan, 151.964),
        "GD": Atom(64, 180, 3.64, "HEX", 234, np.nan, 157.25),
        "TB": Atom(65, 175, 3.6, "HEX", 233, np.nan, 158.92535),
        "DY": Atom(66, 175, 3.59, "HEX", 231, np.nan, 162.5),
        "HO": Atom(67, 175, 3.58, "HEX", 230, np.nan, 164.93033),
        "ER": Atom(68, 175, 3.56, "HEX", 229, np.nan, 167.259),
        "TM": Atom(69, 175, 3.54, "HEX", 227, np.nan, 168.93422),
        "YB": Atom(70, 175, 5.49, "FCC", 226, np.nan, 173.045),
        "LU": Atom(71, 175, 3.51, "HEX", 224, np.nan, 174.9668),
        "HF": Atom(72, 155, 3.2, "HEX", 223, np.nan, 178.49),
        "TA": Atom(73, 145, 3.31, "BCC", 222, np.nan, 180.94788),
        "W": Atom(74, 135, 3.16, "BCC", 218, np.nan, 183.84),
        "RE": Atom(75, 135, 2.76, "HEX", 216, np.nan, 186.207),
        "OS": Atom(76, 130, 2.74, "HEX", 216, np.nan, 190.23),
        "IR": Atom(77, 135, 3.84, "FCC", 213, np.nan, 192.217),
        "PT": Atom(78, 135, 3.92, "FCC", 213, np.nan, 195.084),
        "AU": Atom(79, 135, 4.08, "FCC", 214, np.nan, 196.966569),
        "HG": Atom(80, 150, 2.99, "RHL", 223, np.nan, 200.592),
        "TL": Atom(81, 190, 3.46, "HEX", 196, 190, 204.38),
        "PB": Atom(82, 180, 4.95, "FCC", 202, np.nan, 207.2),
        "BI": Atom(83, 160, 4.75, "RHL", 207, 148, 208.9804),
        "PO": Atom(84, 190, 3.35, "SC", 197, np.nan, 209),
        "AT": Atom(85, np.nan, np.nan, "", 202, np.nan, 210),
        "RN": Atom(86, np.nan, np.nan, "FCC", 220, np.nan, 222),
        "FR": Atom(87, np.nan, np.nan, "BCC", 348, np.nan, 223),
        "RA": Atom(88, 215, np.nan, "", 283, np.nan, 226),
        "AC": Atom(89, 195, 5.31, "FCC", 247, np.nan, 227),
        "TH": Atom(90, 180, 5.08, "FCC", 245, np.nan, 232.0377),
        "PA": Atom(91, 180, 3.92, "TET", 243, np.nan, 231.03588),
        "U": Atom(92, 175, 2.85, "ORC", 241, np.nan, 238.02891),
        "NP": Atom(93, 175, 4.72, "ORC", 239, np.nan, 237),
        "PU": Atom(94, 175, np.nan, "MCL", 243, np.nan, 244),
        "AM": Atom(95, 175, np.nan, "", 244, np.nan, 243),
        "CM": Atom(96, np.nan, np.nan, "", 245, np.nan, 247),
        "BK": Atom(97, np.nan, np.nan, "", 244, np.nan, 247),
        "CF": Atom(98, np.nan, np.nan, "", 245, np.nan, 251),
        "ES": Atom(99, np.nan, np.nan, "", 245, np.nan, 252),
        "FM": Atom(100, np.nan, np.nan, "", 245, np.nan, 257),
        "MD": Atom(101, np.nan, np.nan, "", 246, np.nan, 258),
        "NO": Atom(102, np.nan, np.nan, "", 246, np.nan, 259),
        "LR": Atom(103, np.nan, np.nan, "", 246, np.nan, 262),
        "RF": Atom(104, np.nan, np.nan, "", np.nan, np.nan, 267),
        "DB": Atom(105, np.nan, np.nan, "", np.nan, np.nan, 268),
        "SG": Atom(106, np.nan, np.nan, "", np.nan, np.nan, 271),
        "BH": Atom(107, np.nan, np.nan, "", np.nan, np.nan, 274),
        "HS": Atom(108, np.nan, np.nan, "", np.nan, np.nan, 269),
        "MT": Atom(109, np.nan, np.nan, "", np.nan, np.nan, 276),
        "DS": Atom(110, np.nan, np.nan, "", np.nan, np.nan, 281),
        "RG": Atom(111, np.nan, np.nan, "", np.nan, np.nan, 281),
        "CN": Atom(112, np.nan, np.nan, "", np.nan, np.nan, 285),
        "NH": Atom(113, np.nan, np.nan, "", np.nan, np.nan, 286),
        "FL": Atom(114, np.nan, np.nan, "", np.nan, np.nan, 289),
        "MC": Atom(115, np.nan, np.nan, "", np.nan, np.nan, 288),
        "LV": Atom(116, np.nan, np.nan, "", np.nan, np.nan, 293),
        "TS": Atom(117, np.nan, np.nan, "", np.nan, np.nan, 294),
        "OG": Atom(118, np.nan, np.nan, "", np.nan, np.nan, 294),
    }

    def __getitem__(self, key: str):
        """
        Retrieve a value from the internal data using a given key.

        Parameters
        ----------
        key : str
            The key to use for retrieving the corresponding value from
            the internal data.

        Returns
        -------
        value
            The value associated with the provided key in the internal data.
        """
        return self._elements.get(key, self._default)



""" Strategy pattern to allow for flexible array / FFT backends.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from abc import ABC, abstractmethod
from typing import Tuple, Callable, List
from multiprocessing import shared_memory

from numpy.typing import NDArray
from ..types import ArrayLike, Scalar


class MatchingBackend(ABC):
    """
    A strategy class for template matching backends.

    This class provides an interface to enable users to swap between different
    array and fft implementations while preserving the remaining functionalities
    of the API. The objective is to maintain a numpy like interface that generalizes
    across various backends.

    By delegating attribute access to the provided backend, users can access
    functions/methods of the backend as if they were directly part of this class.

    Parameters
    ----------
    array_backend : object
        The backend object providing array functionalities.
    default_dtype : type
        Data type of real array instances, e.g. np.float32.
    complex_dtype : type
        Data type of complex array instances, e.g. np.complex64.
    default_dtype_int : type
        Data type of integer array instances, e.g. np.int32.

    Attributes
    ----------
    _array_backend : object
        The backend object used to delegate method and attribute calls.
    _default_dtype : type
        Data type of real array instances, e.g. np.float32.
    _complex_dtype : type
        Data type of complex array instances, e.g. np.complex64.
    _default_dtype_int : type
        Data type of integer array instances, e.g. np.int32.

    Examples
    --------
    >>> import numpy as np
    >>> backend = MatchingBackend(
        array_backend = np,
        default_dtype = np.float32,
        complex_dtype = np.complex64,
        default_dtype_int = np.int32
    )
    >>> arr = backend.array([1, 2, 3])
    >>> print(arr)
    [1 2 3]

    Notes
    -----
        Developers should be aware of potential naming conflicts between methods and
        attributes of this class and those of the provided backend.
    """

    def __init__(
        self,
        array_backend,
        default_dtype: type,
        complex_dtype: type,
        default_dtype_int: type,
    ):
        self._array_backend = array_backend
        self._default_dtype = default_dtype
        self._complex_dtype = complex_dtype
        self._default_dtype_int = default_dtype_int

    def __getattr__(self, name: str):
        """
        Delegate attribute access to the underlying backend.

        Parameters
        ----------
        name : str
            The name of the attribute to access.

        Returns
        -------
        attribute
            The attribute from the underlying backend.

        Raises
        ------
        AttributeError
            If the attribute is not found in the backend.
        """
        return getattr(self._array_backend, name)

    def __dir__(self) -> List:
        """
        Return a list of attributes available in this object,
        including those from the backend.

        Returns
        -------
        list
            Sorted list of attributes.
        """
        base_attributes = []
        base_attributes.extend(dir(self.__class__))
        base_attributes.extend(self.__dict__.keys())
        base_attributes.extend(dir(self._array_backend))
        return sorted(base_attributes)

    @staticmethod
    def free_sharedarr(link: shared_memory.SharedMemory):
        """
        Free shared array at link.

        Parameters
        ----------
        link : shared_memory.SharedMemory
            The shared memory link to be freed.
        """
        if type(link) is not shared_memory.SharedMemory:
            return None
        try:
            link.close()
            link.unlink()
        # Shared memory has been freed already
        except FileNotFoundError:
            pass

    @abstractmethod
    def to_backend_array(self, arr: NDArray) -> ArrayLike:
        """
        Convert a numpy array instance to backend array type.

        Parameters
        ----------
        arr : NDArray
            The numpy array instance to be converted.

        Returns
        -------
        ArrayLike
            An array of the specified backend.

        See Also
        --------
        :py:meth:`MatchingBackend.to_cpu_array`
        :py:meth:`MatchingBackend.to_numpy_array`
        """

    @abstractmethod
    def to_numpy_array(self, arr: ArrayLike) -> NDArray:
        """
        Convert an array of given backend to a numpy array.

        Parameters
        ----------
        arr : NDArray
            The array instance to be converted.

        Returns
        -------
        NDArray
            The numpy array equivalent of arr.

        See Also
        --------
        :py:meth:`MatchingBackend.to_cpu_array`
        :py:meth:`MatchingBackend.to_backend_array`
        """

    @abstractmethod
    def to_cpu_array(self, arr: ArrayLike) -> NDArray:
        """
        Convert an array of a given backend to a CPU array of that backend.

        Parameters
        ----------
        arr : NDArray
            The array instance to be converted.

        Returns
        -------
        NDArray
            The numpy array equivalent of arr.

        See Also
        --------
        :py:meth:`MatchingBackend.to_numpy_array`
        :py:meth:`MatchingBackend.to_backend_array`
        """

    @abstractmethod
    def free_cache(self):
        """
        Free cached objects allocated by backend.
        """

    @abstractmethod
    def add(self, arr1: ArrayLike, arr2: ArrayLike, out: ArrayLike = None) -> ArrayLike:
        """
        Interface for element-wise addition of arrays.

        Parameters
        ----------
        arr1 : ArrayLike
            Input array.
        arr2 : ArrayLike
            Input array.
        out : ArrayLike, optional
            Optional output array to store the result. If provided, it must have a shape
            that the inputs broadcast to.

        Returns
        -------
        ArrayLike
            Element-wise sum of the input arrays.
        """

    @abstractmethod
    def subtract(
        self, arr1: ArrayLike, arr2: ArrayLike, out: ArrayLike = None
    ) -> ArrayLike:
        """
        Interface for element-wise subtraction of arrays.

        Parameters
        ----------
        arr1 : ArrayLike
            The minuend array.
        arr2 : ArrayLike
            The subtrahend array.
        out : ArrayLike, optional
            Optional output array to store the result. If provided, it must have a shape
            that the inputs broadcast to.

        Returns
        -------
        ArrayLike
            Element-wise difference of the input arrays.
        """

    @abstractmethod
    def multiply(
        self, arr1: ArrayLike, arr2: ArrayLike, out: ArrayLike = None
    ) -> ArrayLike:
        """
        Interface for element-wise multiplication of arrays.

        Parameters
        ----------
        arr1 : ArrayLike
            Input array.
        arr2 : ArrayLike
            Input array.
        out : ArrayLike, optional
            Optional output array to store the result. If provided, it must have a shape
            that the inputs broadcast to.

        Returns
        -------
        ArrayLike
            Element-wise product of the input arrays.
        """

    @abstractmethod
    def divide(
        self, arr1: ArrayLike, arr2: ArrayLike, out: ArrayLike = None
    ) -> ArrayLike:
        """
        Interface for element-wise division of arrays.

        Parameters
        ----------
        arr1 : ArrayLike
            The numerator array.
        arr2 : ArrayLike
            The denominator array.
        out : ArrayLike, optional
            Optional output array to store the result. If provided, it must have a shape
            that the inputs broadcast to.

        Returns
        -------
        ArrayLike
            Element-wise quotient of the input arrays.
        """

    @abstractmethod
    def mod(self, arr1: ArrayLike, arr2: ArrayLike, out: ArrayLike = None) -> ArrayLike:
        """
        Interface for element-wise modulus of arrays.

        Parameters
        ----------
        arr1 : ArrayLike
            The numerator array.
        arr2 : ArrayLike
            The denominator array.
        out : ArrayLike, optional
            Optional output array to store the result. If provided, it must have a shape
            that the inputs broadcast to.

        Returns
        -------
        ArrayLike
            Element-wise modulus of the input arrays.
        """

    @abstractmethod
    def sum(self, arr: ArrayLike) -> Scalar:
        """
        Compute the sum of array elements.

        Parameters
        ----------
        arr : ArrayLike
            The array whose sum should be computed.

        Returns
        -------
        Scalar
            Sum of the arr.
        """

    @abstractmethod
    def einsum(
        self, arr1: ArrayLike, arr2: ArrayLike, out: ArrayLike = None
    ) -> ArrayLike:
        """
        Interface for einstein notation based summation.

        Parameters
        ----------
        subscripts : str
            Specifies the subscripts for summation as comma separated
            list of subscript label
        arr1 : ArrayLike
            Input array.
        arr2 : ArrayLike
            Input array.
        out : ArrayLike, optional
            Optional output array to store the result. If provided, it must have a shape
            that the inputs broadcast to.

        Returns
        -------
        ArrayLike
            Element-wise sum of the input arrays.
        """

    @abstractmethod
    def mean(self, arr: ArrayLike) -> Scalar:
        """
        Compute the mean of array elements.

        Parameters
        ----------
        arr : ArrayLike
            The array whose mean should be computed.

        Returns
        -------
        Scalar
            Mean value of the arr.
        """

    @abstractmethod
    def std(self, arr: ArrayLike) -> Scalar:
        """
        Compute the standad deviation of array elements.

        Parameters
        ----------
        arr : Scalar
            The array whose standard deviation should be computed.

        Returns
        -------
        Scalar
            The standard deviation of arr.
        """

    @abstractmethod
    def max(self, arr: ArrayLike) -> Scalar:
        """
        Compute the maximum of array elements.

        Parameters
        ----------
        arr : Scalar
            The array whose maximum should be computed.

        Returns
        -------
        Scalar
            The maximum of arr.
        """

    @abstractmethod
    def min(self, arr: ArrayLike) -> Scalar:
        """
        Compute the minimum of array elements.

        Parameters
        ----------
        arr : Scalar
            The array whose maximum should be computed.

        Returns
        -------
        Scalar
            The maximum of arr.
        """

    @abstractmethod
    def maximum(
        self, arr1: ArrayLike, arr2: ArrayLike, out: ArrayLike = None
    ) -> Scalar:
        """
        Compute the element wise maximum of arr1 and arr2.

        Parameters
        ----------
        arr1, arr2 : ArrayLike
            Arrays for which element wise maximum will be computed.
        out : ArrayLike, optional
            Output array.

        Returns
        -------
        ArrayLike
            Element wise maximum of arr1 and arr2.
        """

    @abstractmethod
    def minimum(
        self, arr1: ArrayLike, arr2: ArrayLike, out: ArrayLike = None
    ) -> Scalar:
        """
        Compute the element wise minimum of arr1 and arr2.

        Parameters
        ----------
        arr1, arr2 : ArrayLike
            Arrays for which element wise minimum will be computed.
        out : ArrayLike, optional
            Output array.

        Returns
        -------
        ArrayLike
            Element wise minimum of arr1 and arr2.
        """

    @abstractmethod
    def sqrt(self, arr: ArrayLike) -> ArrayLike:
        """
        Compute the square root of array elements.

        Parameters
        ----------
        arr : ArrayLike
            The array whose square root should be computed.

        Returns
        -------
        ArrayLike
            The squared root of the array.
        """

    @abstractmethod
    def square(self, arr: ArrayLike) -> ArrayLike:
        """
        Compute the square of array elements.

        Parameters
        ----------
        arr : ArrayLike
            The array that should be squared.

        Returns
        -------
        ArrayLike
            The squared arr.
        """

    @abstractmethod
    def abs(self, arr: ArrayLike) -> ArrayLike:
        """
        Compute the absolute of array elements.

        Parameters
        ----------
        arr : ArrayLike
            The array whose absolte should be computed.

        Returns
        -------
        ArrayLike
            The absolute of the array.
        """

    @abstractmethod
    def transpose(self, arr: ArrayLike) -> ArrayLike:
        """
        Compute the transpose of arr.

        Parameters
        ----------
        arr : ArrayLike
            Input array.

        Returns
        -------
        ArrayLike
            The transpose of arr.
        """

    def power(self, *args, **kwargs) -> ArrayLike:
        """
        Compute the n-th power of an array.

        Returns
        -------
        ArrayLike
            The n-th power of the array
        """

    def tobytes(self, arr: ArrayLike) -> str:
        """
        Compute the bytestring representation of arr.

        Parameters
        ----------
        arr : ArrayLike
            Input array.

        Returns
        -------
        str
            Bytestring representation of arr.
        """

    @abstractmethod
    def size(self, arr: ArrayLike) -> int:
        """
        Compute the number of elements of arr.

        Parameters
        ----------
        arr : ArrayLike
            Input array.

        Returns
        -------
        int
            The number of elements in arr.
        """

    @abstractmethod
    def fill(self, arr: ArrayLike, value: float) -> ArrayLike:
        """
        Fill arr with value.

        Parameters
        ----------
        arr : ArrayLike
            The array that should be filled.
        value : float
            The value with which to fill the array.

        Returns
        -------
        ArrayLike
            The array filled with the given value.
        """

    @abstractmethod
    def zeros(self, shape: Tuple[int], dtype: type) -> ArrayLike:
        """
        Returns an aligned array of zeros with specified shape and dtype.

        Parameters
        ----------
        shape : Tuple[int]
            Desired shape for the array.
        dtype : type
            Desired data type for the array.

        Returns
        -------
        ArrayLike
            Byte-aligned array of zeros with specified shape and dtype.
        """

    @abstractmethod
    def full(self, shape: Tuple[int], dtype: type, fill_value: Scalar) -> ArrayLike:
        """
        Returns an aligned array of zeros with specified shape and dtype.

        Parameters
        ----------
        shape : Tuple[int]
            Desired shape for the array.
        dtype : type
            Desired data type for the array.

        Returns
        -------
        ArrayLike
            Byte-aligned array of zeros with specified shape and dtype.
        """

    @abstractmethod
    def eps(self, dtype: type) -> Scalar:
        """
        Returns the eps defined as diffeerence between 1.0 and the next
        representable floating point value larger than 1.0.

        Parameters
        ----------
        dtype : type
            Data type for which eps should be returned.

        Returns
        -------
        Scalar
            The eps for the given data type
        """

    @abstractmethod
    def datatype_bytes(self, dtype: type) -> int:
        """
        Return the number of bytes occupied by a given datatype.

        Parameters
        ----------
        dtype : type
            Datatype for which the number of bytes is to be determined.
            This is typically a data type like `np.float32` or `np.int64`.

        Returns
        -------
        int
            Number of bytes occupied by the datatype.

        Examples
        --------
        >>> MatchingBackend.datatype_bytes(np.float32)
        4
        """

    @abstractmethod
    def clip(
        self, arr: ArrayLike, a_min: Scalar, a_max: Scalar, out: ArrayLike = None
    ) -> ArrayLike:
        """
        Clip elements of arr.

        Parameters
        ----------
        arr : ArrayLike
            Input array.
        a_min : Scalar
            Lower bound.
        a_max : Scalar
            Upper bound.
        out : ArrayLike, optional
            Output array.

        Returns
        -------
        ArrayLike
            Clipped arr.
        """

    @abstractmethod
    def flip(
        self, arr: ArrayLike, a_min: Scalar, a_max: Scalar, out: ArrayLike = None
    ) -> ArrayLike:
        """
        Flip the elements of arr.

        Parameters
        ----------
        arr : ArrayLike
            Input array.

        Returns
        -------
        ArrayLike
            Flipped version of arr.
        """

    @abstractmethod
    def astype(arr: ArrayLike, dtype: type) -> ArrayLike:
        """
        Change the datatype of arr.

        Parameters
        ----------
        arr : ArrayLike
            Input array.
        dtype : type
            Target data type.

        Returns
        -------
        ArrayLike
            Flipped version of arr.
        """

    @abstractmethod
    def arange(self, *args, **kwargs) -> ArrayLike:
        """
        Arange values in increasing order.

        Parameters
        ----------
        arr : ArrayLike
            Input array.
        dtype : type
            Target data type.

        Returns
        -------
        ArrayLike
            Flipped version of arr.
        """

    def stack(self, *args, **kwargs) -> ArrayLike:
        """
        Join a sequence of ArrayLike objects along a specified axis.

        Returns
        -------
        ArrayLike
            The joined input objects.
        """

    @abstractmethod
    def concatenate(self, *args, **kwargs) -> ArrayLike:
        """
        Concatenates arrays along axis.

        Parameters
        ----------
        arr * : ArrayLike
            Input arrays

        Returns
        -------
        ArrayLike
            Concatenated input arrays
        """

    @abstractmethod
    def repeat(self, *args, **kwargs) -> ArrayLike:
        """
        Repeat each array elements after themselves.

        Parameters
        ----------
        arr : ArrayLike
            Input array

        Returns
        -------
        ArrayLike
            Repeated input array
        """

    @abstractmethod
    def topk_indices(self, arr: NDArray, k: int) -> ArrayLike:
        """
        Compute indices of largest elements.

        Parameters
        ----------
        arr : ArrayLike
            Input array.
        dtype : type
            Target data type.

        Returns
        -------
        ArrayLike
            Flipped version of arr.
        """

    def indices(self, *args, **kwargs) -> ArrayLike:
        """
        Creates an array representing the index grid of an input.

        Returns
        -------
        ArrayLike
            The index grid.
        """

    @abstractmethod
    def roll(self, *args, **kwargs) -> ArrayLike:
        """
        Roll array elements along a specified axis.

        Parameters
        ----------
        *args : tuple
            Generic arguments.
        **kwargs : dict
            Generic keyword arguments.

        Returns
        -------
        ArrayLike
            Array with elements rolled.

        See Also
        --------
        numpy.roll : For more detailed documentation on arguments and behavior.

        Examples
        --------
        >>> import numpy as np
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> np.roll(x, 2)
        array([4, 5, 1, 2, 3])
        """

    @abstractmethod
    def unique(self, *args, **kwargs) -> Tuple[ArrayLike]:
        """
        Find the unique elements of an array.

        Parameters
        ----------
        *args : tuple
            Generic arguments.
        **kwargs : dict
            Generic keyword arguments.

        Returns
        -------
        ArrayLike or tuple of ArrayLike
            If `return_index`, `return_inverse`, and `return_counts` keyword
            arguments are all False (the default), this will be an ArrayLike object
            of the sorted unique values. Otherwise, it's a tuple with one
            or more arrays as specified by those keyword arguments.

        See Also
        --------
        numpy.unique : For more detailed documentation on arguments and behavior.

        Examples
        --------
        >>> import numpy as np
        >>> x = np.array([1, 2, 3, 2, 3, 4])
        >>> np.unique(x)
        array([1, 2, 3, 4])
        """

    @abstractmethod
    def argsort(self, *args, **kwargs) -> ArrayLike:
        """
        Perform argsort of arr.

        Parameters
        ----------
        arr : ArrayLike
            Input array.
        dtype : type
            Target data type.

        Returns
        -------
        ArrayLike
            Flipped version of arr.
        """

    @abstractmethod
    def unravel_index(self, indices: ArrayLike, shape: Tuple[int]) -> Tuple[ArrayLike]:
        """
        Convert flat index to array indices.

        Parameters
        ----------
        arr : ArrayLike
            Input array.
        dtype : type
            Target data type.

        Returns
        -------
        ArrayLike
            Flipped version of arr.
        """

    @abstractmethod
    def tril_indices(self, *args, **kwargs) -> ArrayLike:
        """
        Compute indices of upper triangular matrix

        Parameters
        ----------
        arr : ArrayLike
            Input array.
        dtype : type
            Target data type.

        Returns
        -------
        ArrayLike
            Flipped version of arr.
        """

    @abstractmethod
    def max_filter_coordinates(
        self, score_space: ArrayLike, min_distance: Tuple[int]
    ) -> ArrayLike:
        """
        Identifies local maxima in score_space separated by min_distance.

        Parameters
        ----------
        score_space : ArrayLike
            Input score space.
        min_distance : tuple of ints
            Minimum distance along each array axis.

        Returns
        -------
        ArrayLike
            Identified local maxima.
        """

    @abstractmethod
    def preallocate_array(self, shape: Tuple[int], dtype: type) -> ArrayLike:
        """
        Returns an aligned array of zeros with specified shape and dtype.

        Parameters
        ----------
        shape : Tuple[int]
            Desired shape for the array.
        dtype : type
            Desired data type for the array.

        Returns
        -------
        ArrayLike
            Byte-aligned array of zeros with specified shape and dtype.
        """

    @abstractmethod
    def sharedarr_to_arr(
        self, shape: Tuple[int], dtype: str, shm: shared_memory.SharedMemory
    ) -> ArrayLike:
        """
        Returns an array of given shape and dtype from shared memory location.

        Parameters
        ----------
        shape : tuple
            Tuple of integers specifying the shape of the array.
        dtype : str
            String specifying the dtype of the array.
        shm : shared_memory.SharedMemory
            Shared memory object where the array is stored.

        Returns
        -------
        ArrayLike
            Array of the specified shape and dtype from the shared memory location.
        """

    @abstractmethod
    def arr_to_sharedarr(
        self, arr: type, shared_memory_handler: type = None
    ) -> shared_memory.SharedMemory:
        """
        Converts an array to an object shared in memory.

        Parameters
        ----------
        arr : ArrayLike
            Numpy array to convert.
        shared_memory_handler : type, optional
            The type of shared memory handler. Default is None.

        Returns
        -------
        shared_memory.SharedMemory
            The shared memory object containing the numpy array.
        """

    @abstractmethod
    def topleft_pad(
        self, arr: ArrayLike, shape: Tuple[int], padval: int = 0
    ) -> ArrayLike:
        """
        Returns an array that has been padded to a specified shape with a padding
        value at the top-left corner.

        Parameters
        ----------
        arr : ArrayLike
            Input array to be padded.
        shape : Tuple[int]
            Desired shape for the output array.
        padval : int, optional
            Value to use for padding, default is 0.

        Returns
        -------
        ArrayLike
            Array that has been padded to the specified shape.
        """

    @abstractmethod
    def build_fft(
        self,
        fast_shape: Tuple[int],
        fast_ft_shape: Tuple[int],
        real_dtype: type,
        complex_dtype: type,
        **kwargs,
    ) -> Tuple[Callable, Callable]:
        """
        Build forward and inverse real fourier transform functions.

        Parameters
        ----------
        fast_shape : tuple
            Tuple of integers corresponding to fast convolution shape
            (see `compute_convolution_shapes`).
        fast_ft_shape : tuple
            Tuple of integers corresponding to the shape of the fourier
            transform array (see `compute_convolution_shapes`).
        real_dtype : dtype
            Numpy dtype of the inverse fourier transform.
        complex_dtype : dtype
            Numpy dtype of the fourier transform.
        fftargs : dict, optional
            Dictionary passed to pyFFTW builders.
        temp_real : ArrayLike, optional
            Temporary real numpy array, by default None.
        temp_fft : ArrayLike, optional
            Temporary fft numpy array, by default None.

        Returns
        -------
        tuple
            Tuple containing function pointers for forward and inverse real
            fourier transform
        """

    def extract_center(self, arr: ArrayLike, newshape: Tuple[int]) -> ArrayLike:
        """
        Extract the centered portion of an array based on a new shape.

        Parameters
        ----------
        arr : ArrayLike
            Input array.
        newshape : tuple
            Desired shape for the central portion.

        Returns
        -------
        ArrayLike
            Central portion of the array with shape `newshape`.
        """

    @abstractmethod
    def compute_convolution_shapes(
        self, arr1_shape: Tuple[int], arr2_shape: Tuple[int]
    ) -> Tuple[List[int], List[int], List[int]]:
        """
        Computes regular, optimized and fourier convolution shape.

        Parameters
        ----------
        arr1_shape : tuple
            Tuple of integers corresponding to array1 shape.
        arr2_shape : tuple
            Tuple of integers corresponding to array2 shape.

        Returns
        -------
        tuple
            Tuple with regular convolution shape, convolution shape optimized for faster
            fourier transform, shape of the forward fourier transform
            (see :py:meth:`build_fft`).
        """

    @abstractmethod
    def rotate_array(self, *args, **kwargs):
        """
        Perform a rigid transform of arr.

        Parameters
        ----------
        arr : ArrayLike
            Input array.

        Returns
        -------
        ArrayLike
            Transformed version of arr.
        """

    @abstractmethod
    def get_available_memory(self) -> int:
        """
        Returns the available memory available for computations in bytes. For CPU
        operations this corresponds to available RAM. For GPU operations the function
        is expected to return the available GPU memory.
        """

    @abstractmethod
    def reverse(arr: ArrayLike) -> ArrayLike:
        """
        Reverse the order of elements in an array along all its axes.

        Parameters
        ----------
        arr : ArrayLike
            Input array.

        Returns
        -------
        ArrayLike
            Reversed array.
        """

    @abstractmethod
    def set_device(device_index: int):
        """
        Set the active device context and operate as context manager for it
        """

    @abstractmethod
    def device_count() -> int:
        """
        Return the number of available compute devices.
        """



import os
import platform

from setuptools import setup, find_packages
from pybind11.setup_helpers import Pybind11Extension, build_ext

about = {}
with open(os.path.join("tme", "__version__.py"), mode="r") as f:
    exec(f.read(), about)
__version__ = about["__version__"]

if __name__ == "__main__":
    if platform.processor() in ("arm", "i386"):
        c_args = ["-O3", "-ftree-vectorize", "-std=c17"]
        cpp_args = [
            "-O3",
            "-ftree-vectorize",
            "-std=c++11",
            "-funroll-loops",
            "-ffast-math",
        ]
    else:
        c_args = ["-O3", "-march=native", "-std=c17"]
        cpp_args = [
            "-O3",
            "-march=native",
            "-std=c++11",
            "-funroll-loops",
            "-ffast-math",
        ]

    ext_modules = [
        Pybind11Extension(
            "tme.extensions",
            ["src/extensions.cpp"],
            define_macros=[("VERSION_INFO", __version__)],
            extra_compile_args=cpp_args,
        ),
    ]

    package_data = {
        "tme": [
            os.path.join("data", "*.npy"),
            os.path.join("data", "metadata.yaml"),
            os.path.join("data", "scattering_factors.pickle"),
        ]
    }
    setup(
        name="pytme",
        author="Valentin Maurer",
        author_email="valentin.maurer@embl-hamburg.de",
        version=__version__,
        description="Template matching on electron microscopy data.",
        package_data=package_data,
        packages=find_packages(),
        scripts=[
            "scripts/match_template.py",
            "scripts/estimate_ram_usage.py",
            "scripts/preprocessor_gui.py",
            "scripts/preprocess.py",
            "scripts/postprocess.py",
        ],
        ext_modules=ext_modules,
        cmdclass={"build_ext": build_ext},
    )



""" Backend using pytorch and optionally GPU acceleration for
    template matching.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import Tuple, Dict, Callable
from contextlib import contextmanager
from multiprocessing import shared_memory
from multiprocessing.managers import SharedMemoryManager

from numpy.typing import NDArray

from .npfftw_backend import NumpyFFTWBackend
from ..types import NDArray, TorchTensor


class PytorchBackend(NumpyFFTWBackend):
    """
    A pytorch based backend for template matching
    """

    def __init__(
        self,
        device="cuda",
        default_dtype=None,
        complex_dtype=None,
        default_dtype_int=None,
        **kwargs,
    ):
        import torch
        import torch.nn.functional as F

        default_dtype = torch.float32 if default_dtype is None else default_dtype
        complex_dtype = torch.complex64 if complex_dtype is None else complex_dtype
        default_dtype_int = (
            torch.int32 if default_dtype_int is None else default_dtype_int
        )

        super().__init__(
            array_backend=torch,
            default_dtype=default_dtype,
            complex_dtype=complex_dtype,
            default_dtype_int=default_dtype_int,
        )
        self.device = device
        self.F = F
        self._default_dtype_int = torch.int32

    def to_backend_array(self, arr: NDArray) -> TorchTensor:
        if isinstance(arr, self._array_backend.Tensor):
            if arr.device == self.device:
                return arr
            return arr.to(self.device)
        return self.tensor(arr, device=self.device)

    def to_numpy_array(self, arr: TorchTensor) -> NDArray:
        return arr.cpu().numpy()

    def to_cpu_array(self, arr: TorchTensor) -> NDArray:
        return arr.cpu()

    def free_cache(self):
        self._array_backend.cuda.empty_cache()

    def mod(self, x1, x2, *args, **kwargs):
        x1 = self.to_backend_array(x1)
        x2 = self.to_backend_array(x2)
        return self._array_backend.remainder(x1, x2, *args, **kwargs)

    def sum(self, *args, **kwargs) -> NDArray:
        return self._array_backend.sum(*args, **kwargs)

    def mean(self, *args, **kwargs) -> NDArray:
        return self._array_backend.mean(*args, **kwargs)

    def std(self, *args, **kwargs) -> NDArray:
        return self._array_backend.std(*args, **kwargs)

    def max(self, *args, **kwargs) -> NDArray:
        ret = self._array_backend.amax(*args, **kwargs)
        if type(ret) == self._array_backend.Tensor:
            return ret
        return ret[0]

    def min(self, *args, **kwargs) -> NDArray:
        ret = self._array_backend.amin(*args, **kwargs)
        if type(ret) == self._array_backend.Tensor:
            return ret
        return ret[0]

    def maximum(self, x1, x2, *args, **kwargs) -> NDArray:
        x1 = self.to_backend_array(x1)
        x2 = self.to_backend_array(x2)
        return self._array_backend.maximum(input=x1, other=x2, *args, **kwargs)

    def minimum(self, x1, x2, *args, **kwargs) -> NDArray:
        x1 = self.to_backend_array(x1)
        x2 = self.to_backend_array(x2)
        return self._array_backend.minimum(input=x1, other=x2, *args, **kwargs)

    def tobytes(self, arr):
        return arr.cpu().numpy().tobytes()

    def size(self, arr):
        return arr.numel()

    def zeros(self, shape, dtype=None):
        return self._array_backend.zeros(shape, dtype=dtype, device=self.device)

    def preallocate_array(self, shape: Tuple[int], dtype: type) -> NDArray:
        """
        Returns a byte-aligned array of zeros with specified shape and dtype.

        Parameters
        ----------
        shape : Tuple[int]
            Desired shape for the array.
        dtype : type
            Desired data type for the array.

        Returns
        -------
        NDArray
            Byte-aligned array of zeros with specified shape and dtype.
        """
        arr = self._array_backend.zeros(shape, dtype=dtype, device=self.device)
        return arr

    def full(self, shape, fill_value, dtype=None):
        return self._array_backend.full(
            shape, dtype=dtype, fill_value=fill_value, device=self.device
        )

    def datatype_bytes(self, dtype: type) -> int:
        temp = self.zeros(1, dtype=dtype)
        return temp.element_size()

    def fill(self, arr: TorchTensor, value: float):
        arr.fill_(value)

    def astype(self, arr, dtype):
        return arr.to(dtype)

    def flip(self, a, axis, **kwargs):
        return self._array_backend.flip(input=a, dims=axis, **kwargs)

    def arange(self, *args, **kwargs):
        return self._array_backend.arange(*args, **kwargs, device=self.device)

    def stack(self, *args, **kwargs):
        return self._array_backend.stack(*args, **kwargs)

    def topk_indices(self, arr, k):
        temp = arr.reshape(-1)
        values, indices = self._array_backend.topk(temp, k)
        indices = self.unravel_index(indices=indices, shape=arr.shape)
        return indices

    def indices(self, shape: Tuple[int]) -> TorchTensor:
        grids = [self.arange(x) for x in shape]
        mesh = self._array_backend.meshgrid(*grids, indexing="ij")
        return self._array_backend.stack(mesh)

    def unravel_index(self, indices, shape):
        indices = self.to_backend_array(indices)
        shape = self.to_backend_array(shape)
        strides = self._array_backend.cumprod(shape.flip(0), dim=0).flip(0)
        strides = self._array_backend.cat(
            (strides[1:], self.to_backend_array([1])),
        )
        unraveled_coords = (indices.view(-1, 1) // strides.view(1, -1)) % shape.view(
            1, -1
        )
        if unraveled_coords.size(0) == 1:
            return tuple(unraveled_coords[0, :].tolist())

        else:
            return tuple(unraveled_coords.T)

    def roll(self, a, shift, axis, **kwargs):
        shift = tuple(shift)
        return self._array_backend.roll(input=a, shifts=shift, dims=axis, **kwargs)

    def unique(
        self,
        ar,
        return_index: bool = False,
        return_inverse: bool = False,
        return_counts: bool = False,
        axis: int = None,
        sorted: bool = True,
    ):
        # https://github.com/pytorch/pytorch/issues/36748#issuecomment-1478913448
        unique, inverse, counts = self._array_backend.unique(
            ar, return_inverse=True, return_counts=True, dim=axis, sorted=sorted
        )
        inverse = inverse.reshape(-1)

        if return_index:
            inv_sorted = inverse.argsort(stable=True)
            tot_counts = self._array_backend.cat(
                (counts.new_zeros(1), counts.cumsum(dim=0))
            )[:-1]
            index = inv_sorted[tot_counts]

        ret = unique
        if return_index or return_inverse or return_counts:
            ret = [unique]

        if return_index:
            ret.append(index)
        if return_inverse:
            ret.append(inverse)
        if return_counts:
            ret.append(counts)

        return ret

    def max_filter_coordinates(self, score_space, min_distance: Tuple[int]):
        if score_space.ndim == 3:
            func = self._array_backend.nn.MaxPool3d
        elif score_space.ndim == 2:
            func = self._array_backend.nn.MaxPool2d
        else:
            raise NotImplementedError("Operation only implemented for 2 and 3D inputs.")

        pool = func(kernel_size=min_distance, return_indices=True)
        _, indices = pool(score_space.reshape(1, 1, *score_space.shape))
        coordinates = self.unravel_index(indices.reshape(-1), score_space.shape)
        coordinates = self.transpose(self.stack(coordinates))
        return coordinates

    def repeat(self, *args, **kwargs):
        return self._array_backend.repeat_interleave(*args, **kwargs)

    def sharedarr_to_arr(
        self, shape: Tuple[int], dtype: str, shm: TorchTensor
    ) -> TorchTensor:
        if self.device == "cuda":
            return shm

        required_size = int(self._array_backend.prod(self.to_backend_array(shape)))

        ret = self._array_backend.frombuffer(shm.buf, dtype=dtype)[
            :required_size
        ].reshape(shape)
        return ret

    def arr_to_sharedarr(
        self, arr: TorchTensor, shared_memory_handler: type = None
    ) -> TorchTensor:
        if self.device == "cuda":
            return arr

        nbytes = arr.numel() * arr.element_size()

        if type(shared_memory_handler) == SharedMemoryManager:
            shm = shared_memory_handler.SharedMemory(size=nbytes)
        else:
            shm = shared_memory.SharedMemory(create=True, size=nbytes)

        shm.buf[:nbytes] = arr.numpy().tobytes()

        return shm

    def transpose(self, arr):
        return arr.permute(*self._array_backend.arange(arr.ndim - 1, -1, -1))

    def power(self, *args, **kwargs):
        return self._array_backend.pow(*args, **kwargs)

    def rotate_array(
        self,
        arr: TorchTensor,
        rotation_matrix: TorchTensor,
        arr_mask: TorchTensor = None,
        translation: TorchTensor = None,
        out: TorchTensor = None,
        out_mask: TorchTensor = None,
        order: int = 1,
        **kwargs,
    ):
        """
        Rotates the given tensor `arr` based on the provided `rotation_matrix`.

        This function optionally allows for rotating an accompanying mask
        tensor (`arr_mask`) alongside the main tensor. The rotation is defined
        by the `rotation_matrix` and the optional `translation`.

        Parameters
        ----------
        arr : TorchTensor
            The input tensor to be rotated.
        rotation_matrix : TorchTensor
            The rotation matrix to apply. Must be square and of shape [d x d].
        arr_mask : TorchTensor, optional
            The mask of `arr` to be equivalently rotated.
        translation : TorchTensor, optional
            The translation to apply after rotation. Shape should
            match tensor dimensions [d].
        out : TorchTensor, optional
            The output tensor to hold the rotated `arr`. If not provided, a new
            tensor will be created.
        out_mask : TorchTensor, optional
            The output tensor to hold the rotated `arr_mask`. If not provided and
            `arr_mask` is given, a new tensor will be created.
        order : int, optional
            Spline interpolation order. Supports orders:

            +-------+---------------------------------------------------------------+
            |   0   | Use 'nearest' neighbor interpolation.                         |
            +-------+---------------------------------------------------------------+
            |   1   | Use 'bilinear' interpolation for smoother results.            |
            +-------+---------------------------------------------------------------+
            |   3   | Use 'bicubic' interpolation for higher order smoothness.      |
            +-------+---------------------------------------------------------------+

        Returns
        -------
        out, out_mask : TorchTensor or Tuple[TorchTensor, TorchTensor] or None
            Returns the rotated tensor(s). If `out` and `out_mask` are provided, the
            function will return `None`.
            If only `arr` is provided without `out`, it returns rotated `arr`.
            If both `arr` and `arr_mask` are given without `out` and `out_mask`, it
            returns a tuple of rotated tensors.

        Notes
        -----
        Only a region of the size of `arr` and `arr_mask` is considered for
        interpolation in `out` and `out_mask` respectively.

        Currently bicubic interpolation is not supported for 3D inputs.
        """
        device = arr.device
        mode_mapping = {0: "nearest", 1: "bilinear", 3: "bicubic"}
        mode = mode_mapping.get(order, None)
        if mode is None:
            modes = ", ".join([str(x) for x in mode_mapping.keys()])
            raise ValueError(
                f"Got {order} but supported interpolation orders are: {modes}."
            )
        rotate_mask = arr_mask is not None
        return_type = (out is None) + 2 * rotate_mask * (out_mask is None)

        out = self.zeros_like(arr) if out is None else out
        if translation is None:
            translation = self._array_backend.zeros(arr.ndim, device=device)

        normalized_translation = self.divide(
            -2.0 * translation, self.tensor(arr.shape, device=arr.device)
        )

        rotation_matrix_pull = self.linalg.inv(self.flip(rotation_matrix, [0, 1]))

        out_slice = tuple(slice(0, x) for x in arr.shape)
        out[out_slice] = self._affine_transform(
            arr=arr,
            rotation_matrix=rotation_matrix_pull,
            translation=normalized_translation,
            mode=mode,
        )

        if rotate_mask:
            out_mask_slice = tuple(slice(0, x) for x in arr_mask.shape)
            if out_mask is None:
                out_mask = self._array_backend.zeros_like(arr_mask)
            out_mask[out_mask_slice] = self._affine_transform(
                arr=arr_mask,
                rotation_matrix=rotation_matrix_pull,
                translation=normalized_translation,
                mode=mode,
            )

        match return_type:
            case 0:
                return None
            case 1:
                return out
            case 2:
                return out_mask
            case 3:
                return out, out_mask

    def build_fft(
        self, fast_shape: Tuple[int], fast_ft_shape: Tuple[int], **kwargs
    ) -> Tuple[Callable, Callable]:
        """
        Build fft builder functions.

        Parameters
        ----------
        fast_shape : tuple
            Tuple of integers corresponding to fast convolution shape
            (see `compute_convolution_shapes`).
        fast_ft_shape : tuple
            Tuple of integers corresponding to the shape of the fourier
            transform array (see `compute_convolution_shapes`).
        **kwargs : dict, optional
            Additional parameters that are not used for now.

        Returns
        -------
        tuple
            Tupple containing callable rfft and irfft object.
        """

        def rfftn(
            arr: TorchTensor, out: TorchTensor, shape: Tuple[int] = fast_shape
        ) -> None:
            return self._array_backend.fft.rfftn(arr, s=shape, out=out)

        def irfftn(
            arr: TorchTensor, out: TorchTensor, shape: Tuple[int] = fast_shape
        ) -> None:
            return self._array_backend.fft.irfftn(arr, s=shape, out=out)

        return rfftn, irfftn

    def _affine_transform(
        self,
        arr: TorchTensor,
        rotation_matrix: TorchTensor,
        translation: TorchTensor,
        mode,
    ) -> TorchTensor:
        """
        Performs an affine transformation on the given tensor.

        The affine transformation is defined by the provided `rotation_matrix`
        and the `translation` vector. The transformation is applied to the
        input tensor `arr`.

        Parameters
        ----------
        arr : TorchTensor
            The input tensor on which the transformation will be applied.
        rotation_matrix : TorchTensor
            The matrix defining the rotation component of the transformation.
        translation : TorchTensor
            The vector defining the translation to be applied post rotation.
        mode : str
            Interpolation mode to use. Options are: 'nearest', 'bilinear', 'bicubic'.

        Returns
        -------
        TorchTensor
            The tensor after applying the affine transformation.
        """

        transformation_matrix = self._array_backend.zeros(
            arr.ndim, arr.ndim + 1, device=arr.device, dtype=arr.dtype
        )
        transformation_matrix[:, : arr.ndim] = rotation_matrix
        transformation_matrix[:, arr.ndim] = translation

        size = self.Size([1, 1, *arr.shape])
        grid = self.F.affine_grid(
            theta=transformation_matrix.unsqueeze(0), size=size, align_corners=False
        )
        output = self.F.grid_sample(
            input=arr.unsqueeze(0).unsqueeze(0),
            grid=grid,
            mode=mode,
            align_corners=False,
        )

        return output.squeeze()

    def get_available_memory(self) -> int:
        if self.device == "cpu":
            return super().get_available_memory()
        return self._array_backend.cuda.mem_get_info()[0]

    @contextmanager
    def set_device(self, device_index: int):
        """
        Set the active GPU device as a context.

        This method sets the active GPU device for operations within the context.

        Parameters
        ----------
        device_index : int
            Index of the GPU device to be set as active.

        Yields
        ------
        None
            Operates as a context manager, yielding None and providing
            the set GPU context for enclosed operations.
        """
        if self.device == "cuda":
            with self._array_backend.cuda.device(device_index):
                yield
        else:
            yield None

    def device_count(self) -> int:
        """
        Return the number of available GPU devices.

        Returns
        -------
        int
            Number of available GPU devices.
        """
        return self._array_backend.cuda.device_count()

    def reverse(self, arr: TorchTensor) -> TorchTensor:
        """
        Reverse the order of elements in a tensor along all its axes.

        Parameters
        ----------
        tensor : TorchTensor
            Input tensor.

        Returns
        -------
        TorchTensor
            Reversed tensor.
        """
        return self._array_backend.flip(arr, [i for i in range(arr.ndim)])



#!python3

from os import listdir
from os.path import dirname, join, basename

import numpy as np
import yaml


def quat_to_numpy(filepath):
    data = []
    with open(join(filepath, filepath), "r", encoding="utf-8") as ifile:
        for line in ifile:
            if line.startswith("#") or line.startswith("format quaternion"):
                continue
            line_split = line.strip().split()
            if len(line_split) == 3:
                n, angle, c = line_split
                continue
            data.append(line_split)

    data = np.array(data).astype(float)
    return data, int(n), float(angle), float(c)


if __name__ == "__main__":
    current_directory = dirname(__file__)
    files = listdir(current_directory)

    files = [file for file in files if file.endswith(".quat")]
    files = [join(current_directory, file) for file in files]
    numpy_names = [
        join(current_directory, file.replace("quat", "npy")) for file in files
    ]

    metadata = {}
    for file, np_out in zip(files, numpy_names):
        quaternions, n, angle, c = quat_to_numpy(file)
        np.save(np_out, quaternions)
        metadata[basename(np_out)] = [n, angle, c]
    with open(join(current_directory, "metadata.yaml"), "w", encoding="utf-8") as ofile:
        yaml.dump(metadata, ofile, default_flow_style=False)



""" Implements parsers for atomic structure file formats.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""
import re
from collections import deque
from typing import List, Dict
from abc import ABC, abstractmethod

import numpy as np


class Parser(ABC):
    """
    Base class for structure file parsers.

    Classes inheriting from :py:class:`Parser` need to define
    a ``parse_input`` method that accepts a list of lines and returns a
    dictionary representation of the data.
    """

    def __init__(self, filename: str, mode: str = "r") -> None:
        """
        Initialize a Parser object.

        Parameters
        ----------
        filename : str
            File name to parse data from.

        mode : str, optional
            Mode to open the file. Default is 'r' for read.
        """
        with open(filename, "r") as infile:
            data = infile.read()

        data = deque(filter(lambda line: line and line[0] != "#", data.split("\n")))
        self._data = self.parse_input(data)

    def __getitem__(self, key: str):
        """
        Retrieve a value from the internal data using a given key.

        Parameters
        ----------
        key : str
            The key to use for retrieving the corresponding value from
            the internal data.

        Returns
        -------
        value
            The value associated with the provided key in the internal data.
        """
        return self._data[key]

    def __contains__(self, key) -> bool:
        """
        Check if a given key exists in the internal data.

        Parameters
        ----------
        key : str
            The key to check for in the internal data.

        Returns
        -------
        bool
            True if the key exists in the internal data, False otherwise.
        """
        return key in self._data

    def get(self, key, default):
        """
        Retrieve a value from the internal data using a given key. If the
        key does not exist, return a default value.

        Parameters
        ----------
        key : str
            The key to use for retrieving the corresponding value from
             the internal data.

        default : Any
            The value to return if the key does not exist in the internal data.

        Returns
        -------
        value
            The value associated with the provided key in the internal data,
            or the default value if the key does not exist.
        """
        if key in self._data:
            return self[key]
        return default

    def keys(self):
        """
        List keys available in internal dictionary.
        """
        return self._data.keys()

    def values(self):
        """
        List values available in internal dictionary.
        """
        return self._data.values()

    def items(self):
        """
        List items available in internal dictionary.
        """
        return self._data.items()

    @abstractmethod
    def parse_input(self, lines: List[str]) -> Dict:
        """
        Parse a list of lines from a file and convert the data into a dictionary.

        This function is not intended to be called directly, but should rather be
        defined by classes inheriting from :py:class:`Parser` to parse a given
        file format.

        Parameters
        ----------
        lines : list of str
            The lines of a structure file to parse.

        Returns
        -------
        dict
            A dictionary containing the parsed data.
        """


class PDBParser(Parser):
    """
    A Parser subclass for converting PDB file data into a dictionary representation.
    This class is specifically designed to work with PDB file format.

    References
    ----------
    .. [1]  https://www.cgl.ucsf.edu/chimera/docs/UsersGuide/tutorials/pdbintro.html
    """

    def parse_input(self, lines: List[str]) -> Dict:
        """
        Parse a list of lines from a PDB file and convert the data into a dictionary.

        Parameters
        ----------
        lines : list of str
            The lines of a PDB file to parse.

        Returns
        -------
        dict
            A dictionary containing the parsed data from the PDB file.
        """
        metadata = {
            "resolution": re.compile(
                r"(.)+?(EFFECTIVE RESOLUTION\s+\(ANGSTROMS\)){1}(.)+?(\d+\.\d+)(\s)*$"
            ),
            "reconstruction_method": re.compile(
                r"(.)+?(RECONSTRUCTION METHOD)+(.)+?(\w+\s*\w+)(\s)*$"
            ),
            "electron_source": re.compile(r"(.)+?(SOURCE)+(.)+?(\w+\s*\w+)(\s)*$"),
            "illumination_mode": re.compile(
                r"(.)+?(ILLUMINATION MODE)+(.)+?(\w+\s*\w+)(\s)*$"
            ),
            "microscope_mode": re.compile(
                r"(.)+?(IMAGING MODE)+(.)+?(\w+\s*\w+)(\s)*$"
            ),
            "microscope_model": re.compile(
                r"(.)+?(MICROSCOPE MODEL)+(.+?:\s+)+?(.+)(\s)*$"
            ),
        }

        data = {
            "record_type": [],
            "atom_serial_number": [],
            "atom_name": [],
            "alternate_location_indicator": [],
            "residue_name": [],
            "chain_identifier": [],
            "residue_sequence_number": [],
            "code_for_residue_insertion": [],
            "atom_coordinate": [],
            "occupancy": [],
            "temperature_factor": [],
            "segment_identifier": [],
            "element_symbol": [],
            "charge": [],
            "details": {},
        }
        data["details"]["resolution"] = np.nan

        for line in lines:
            if line.startswith("REMARK"):
                matches = [(key, metadata[key].match(line)) for key in metadata]
                matches = [match for match in matches if match[1]]
                for key, match in matches:
                    data["details"][key] = match.group(4)
                    _ = metadata.pop(key)
            elif line.startswith("ATOM") or line.startswith("HETATM"):
                data["record_type"].append(line[0:6])
                data["atom_serial_number"].append(line[6:11])
                data["atom_name"].append(line[12:16])
                data["alternate_location_indicator"].append(line[16])
                data["residue_name"].append(line[17:20])

                data["chain_identifier"].append(line[21])
                data["residue_sequence_number"].append(line[22:26])
                data["code_for_residue_insertion"].append(line[26])
                data["atom_coordinate"].append((line[30:38], line[38:46], line[46:54]))
                data["occupancy"].append(line[54:60])
                data["temperature_factor"].append(line[60:66])
                data["segment_identifier"].append(line[74:76])
                data["element_symbol"].append(line[76:78])
                data["charge"].append(line[78:80])

        data["details"]["resolution"] = float(data["details"]["resolution"])

        return data


class MMCIFParser(Parser):
    """
    A Parser subclass for converting MMCIF file data into a dictionary representation.
    This implementation heavily relies on the atomium library:

    References
    ----------
    .. [1]  Ireland, S. M., & Martin, A. C. R. (2020). atomium (Version 1.0.0)
            [Computer software]. https://doi.org/10.1093/bioinformatics/btaa072
    """

    def parse_input(self, lines: List[str]) -> Dict:
        """
        Parse a list of lines from an MMCIF file and convert the data into a dictionary.

        Parameters
        ----------
        lines : list of str
            The lines of an MMCIF file to parse.

        Returns
        -------
        dict
            A dictionary containing the parsed data from the MMCIF file.
        """
        lines = self._consolidate_strings(lines)
        blocks = self._split_in_blocks(lines)
        mmcif_dict = {}
        for block in blocks:
            if block["lines"][0] == "loop_":
                mmcif_dict[block["category"]] = self._loop_block_to_dict(block)
            else:
                mmcif_dict[block["category"]] = self._non_loop_block_to_dict(block)
        return mmcif_dict

    @staticmethod
    def _consolidate_strings(lines: List[str]) -> List[str]:
        """
        Consolidate multi-line strings that have been separated by semicolons in a
        list of strings.

        Parameters
        ----------
        lines : deque of str
            Deque of strings where each string is a line from an MMCIF file.

        Returns
        -------
        deque of str
            A deque of consolidated strings from the given input.
        """
        new_lines = deque()
        while lines:
            line = lines.popleft()
            if line.startswith(";"):
                string = [line[1:].strip()]
                while not lines[0].startswith(";"):
                    string.append(lines.popleft())
                lines.popleft()
                new_lines[-1] += ' "{}"'.format(
                    " ".join(string).replace('"', "").replace("'", "'")
                )
            else:
                new_lines.append(line.replace('"', "").replace("'", "'"))
        return new_lines

    @staticmethod
    def _split_in_blocks(lines: List[str]) -> List[Dict]:
        """
        Split a deque of consolidated strings into a list of dictionaries,
        each representing a block of data.

        Parameters
        ----------
        lines : deque of str
            Deque of consolidated strings where each string is a line from
            an MMCIF file.

        Returns
        -------
        list of dict
            A list of dictionaries where each dictionary represents a block
             of data from the MMCIF file.
        """
        category = None
        block, blocks = [], []
        while lines:
            line = lines.popleft()
            if line.startswith("data_"):
                continue
            if line.startswith("_"):
                line_category = line.split(".")[0]
                if line_category != category:
                    if category:
                        blocks.append({"category": category[1:], "lines": block})
                    category = line_category
                    block = []
            if line.startswith("loop_"):
                if category:
                    blocks.append({"category": category[1:], "lines": block})
                category = lines[0].split(".")[0]
                block = []
            block.append(line)
        if block:
            blocks.append({"category": category[1:], "lines": block})
        return blocks

    @staticmethod
    def _non_loop_block_to_dict(block: Dict) -> Dict:
        """
        Convert a non-loop block of data into a dictionary.

        Parameters
        ----------
        block : dict
            A dictionary representing a non-loop block of data from an MMCIF file.

        Returns
        -------
        dict
            A dictionary representing the parsed data from the given non-loop block.
        """
        d = {}
        # category = block["lines"][0].split(".")[0]
        for index in range(len(block["lines"]) - 1):
            if block["lines"][index + 1][0] != "_":
                block["lines"][index] += " " + block["lines"][index + 1]
        block["lines"] = [line for line in block["lines"] if line[0] == "_"]
        for line in block["lines"]:
            name = line.split(".")[1].split()[0]
            value = " ".join(line.split()[1:])
            d[name] = value
        return d

    def _loop_block_to_dict(self, block: Dict) -> Dict:
        """
        Convert a loop block of data into a dictionary.

        Parameters
        ----------
        block : dict
            A dictionary representing a loop block of data from an MMCIF file.

        Returns
        -------
        dict
            A dictionary representing the parsed data from the given loop block.
        """
        names, lines = [], []
        body_start = 0
        for index, line in enumerate(block["lines"][1:], start=1):
            if not line.startswith("_" + block["category"]):
                body_start = index
                break
        names = [line.split(".")[1].rstrip() for line in block["lines"][1:body_start]]
        lines = [self._split_line(line) for line in block["lines"][body_start:]]
        # reunites broken lines
        for n in range(len(lines) - 1):
            while n < len(lines) - 1 and len(lines[n]) + len(lines[n + 1]) <= len(
                names
            ):
                lines[n] += lines.pop(n + 1)
        res = {name: [] for name in names}
        for line in lines:
            for name, value in zip(names, line):
                res[name].append(value)
        return res

    @staticmethod
    def _split_line(line: str) -> List[str]:
        """
        Split a string into substrings, ignoring quotation marks within the string.

        Parameters
        ----------
        line : str
            The string to be split.

        Returns
        -------
        list of str
            A list of substrings resulting from the split operation on the given string.
        """
        if not re.search("['\"]", line):
            return line.split()

        chars = deque(line.strip())
        values, value, in_string = [], [], False
        while chars:
            char = chars.popleft()
            if char == " " and not in_string:
                values.append("".join(value))
                value = []
            elif char == '"':
                in_string = not in_string
                value.append(char)
            else:
                value.append(char)

        values.append(value)
        return ["".join(v) for v in values if v]



""" Data class for holding template matching data.

    Copyright (c) 2023 European Molecular Biology Laboratory

    Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import Tuple, List

import numpy as np
from numpy.typing import NDArray

from . import Density
from .backends import backend
from .matching_utils import compute_full_convolution_index


class MatchingData:
    """
    Contains data required for template matching.

    Parameters
    ----------
    target : np.ndarray or Density
        Target data array for template matching.
    template : np.ndarray or Density
        Template data array for template matching.

    """

    def __init__(self, target: NDArray, template: NDArray):
        self._default_dtype = np.float32
        self._complex_dtype = np.complex64

        self._target = target
        self._target_mask = None
        self._template_mask = None
        self._translation_offset = np.zeros(len(target.shape), dtype=int)

        self.template = template

        self._target_pad = np.zeros(len(target.shape), dtype=int)
        self._template_pad = np.zeros(len(template.shape), dtype=int)

        self.template_filter = {}
        self.target_filter = {}

        self._invert_target = False

    @staticmethod
    def _shape_to_slice(shape: Tuple[int]):
        return tuple(slice(0, dim) for dim in shape)

    @classmethod
    def _slice_to_mesh(cls, slice_variable: (slice,), shape: (int,)):
        if slice_variable is None:
            slice_variable = cls._shape_to_slice(shape)
        ranges = [range(slc.start, slc.stop) for slc in slice_variable]
        indices = np.meshgrid(*ranges, sparse=True, indexing="ij")
        return indices

    @staticmethod
    def _load_array(arr: NDArray):
        """
        Load ``arr``,  If ``arr`` type is memmap, reload from disk.

        Parameters
        ----------
        arr : NDArray
            Array to load.

        Returns
        -------
        NDArray
            Loaded array.
        """

        if type(arr) == np.memmap:
            return np.memmap(arr.filename, mode="r", shape=arr.shape, dtype=arr.dtype)
        return arr

    def subset_array(
        self, arr: NDArray, arr_slice: Tuple[slice], padding: NDArray
    ) -> NDArray:
        """
        Extract a subset of the input array according to the given slice and
        apply padding.

        Parameters
        ----------
        arr : NDArray
            The input array from which a subset is extracted.
        arr_slice : tuple of slice
            Defines the region of the input array to be extracted.
        padding : NDArray
            Padding values for each dimension. If the padding exceeds the array
            dimensions, the extra regions are filled with the mean of the array
            values, otherwise, the
            values in ``arr`` are used.

        Returns
        -------
        NDArray
            Subset of the input array with padding applied.
        """
        padding = np.maximum(padding, 0)

        slice_start = np.array([x.start for x in arr_slice], dtype=int)
        slice_stop = np.array([x.stop for x in arr_slice], dtype=int)
        slice_shape = np.subtract(slice_stop, slice_start)

        padding = np.add(padding, np.mod(padding, 2))
        left_pad = right_pad = np.divide(padding, 2).astype(int)

        data_voxels_left = np.minimum(slice_start, left_pad)
        data_voxels_right = np.minimum(
            np.subtract(arr.shape, slice_stop), right_pad
        ).astype(int)

        ret_shape = np.add(slice_shape, padding)
        arr_start = np.subtract(slice_start, data_voxels_left)
        arr_stop = np.add(slice_stop, data_voxels_right)
        arr_slice = tuple(slice(*pos) for pos in zip(arr_start, arr_stop))
        arr_mesh = self._slice_to_mesh(arr_slice, arr.shape)

        subset_start = np.subtract(left_pad, data_voxels_left)
        subset_stop = np.add(subset_start, np.subtract(arr_stop, arr_start))
        subset_slice = tuple(slice(*prod) for prod in zip(subset_start, subset_stop))
        subset_mesh = self._slice_to_mesh(subset_slice, ret_shape)

        if type(arr) == Density:
            if type(arr.data) == np.memmap:
                arr = Density.from_file(arr.data.filename, subset=arr_slice).data
            else:
                arr = np.asarray(arr.data[*arr_mesh])
        else:
            if type(arr) == np.memmap:
                arr = np.memmap(
                    arr.filename, mode="r", shape=arr.shape, dtype=arr.dtype
                )
            arr = np.asarray(arr[*arr_mesh])
        ret = np.full(
            shape=np.add(slice_shape, padding), fill_value=arr.mean(), dtype=arr.dtype
        )
        ret[*subset_mesh] = arr

        return ret

    def subset_by_slice(
        self,
        target_slice: Tuple[slice] = None,
        template_slice: Tuple[slice] = None,
        target_pad: NDArray = None,
        template_pad: NDArray = None,
        invert_target: bool = False,
    ) -> "MatchingData":
        """
        Slice the instance arrays based on the provided slices.

        Parameters
        ----------
        target_slice : tuple of slice, optional
            Slices for the target. If not provided, the full shape is used.
        template_slice : tuple of slice, optional
            Slices for the template. If not provided, the full shape is used.
        target_pad : NDArray, optional
            Padding for target. Defaults to zeros. If padding exceeds target,
            pad with mean.
        template_pad : NDArray, optional
            Padding for template. Defaults to zeros. If padding exceeds template,
            pad with mean.

        Returns
        -------
        MatchingData
            Newly allocated sliced class instance.
        """
        target_shape = self.target.shape
        template_shape = self._template.shape

        if target_slice is None:
            target_slice = self._shape_to_slice(target_shape)
        if template_slice is None:
            template_slice = self._shape_to_slice(template_shape)

        if target_pad is None:
            target_pad = np.zeros(len(self.target.shape), dtype=int)
        if template_pad is None:
            template_pad = np.zeros(len(self.target.shape), dtype=int)

        indices = compute_full_convolution_index(
            outer_shape=self._target.shape,
            inner_shape=self._template.shape,
            outer_split=target_slice,
            inner_split=template_slice,
        )

        target_subset = self.subset_array(
            arr=self._target, arr_slice=target_slice, padding=target_pad
        )
        if self._invert_target:
            target_subset *= -1
            target_min, target_max = target_subset.min(), target_subset.max()
            target_subset = (target_subset - target_min) / (target_max - target_min)
        template_subset = self.subset_array(
            arr=self._template,
            arr_slice=template_slice,
            padding=template_pad,
        )
        ret = self.__class__(target=target_subset, template=template_subset)

        ret._translation_offset = np.add(
            [x.start for x in target_slice],
            [x.start for x in template_slice],
        )
        ret.template_filter = self.template_filter

        ret.rotations, ret.indices = self.rotations, indices
        ret._target_pad, ret._template_pad = target_pad, template_pad
        ret._invert_target = self._invert_target

        if self._target_mask is not None:
            ret.target_mask = self.subset_array(
                arr=self._target_mask, arr_slice=target_slice, padding=target_pad
            )
        if self._template_mask is not None:
            ret.template_mask = self.subset_array(
                arr=self._template_mask,
                arr_slice=template_slice,
                padding=template_pad,
            )

        return ret

    def to_backend(self) -> None:
        """
        Transfer the class instance's numpy arrays to the current backend.
        """
        for attr_name, attr_value in vars(self).items():
            if isinstance(attr_value, np.ndarray):
                converted_array = backend.to_backend_array(attr_value.copy())
                setattr(self, attr_name, converted_array)

        self._default_dtype = backend._default_dtype
        self._complex_dtype = backend._complex_dtype

    @property
    def rotations(self):
        """Return rotation matrices used for fitting."""
        return self._rotations

    @rotations.setter
    def rotations(self, rotations: NDArray):
        """
        Set and reshape the rotation matrices for fitting.

        Parameters
        ----------
        rotations : NDArray
            Rotations in shape (3 x 3), (1 x 3 x 3), or (n x k x k).
        """
        if rotations.__class__ != np.ndarray:
            raise ValueError("Rotation set has to be of type numpy ndarray.")
        if rotations.ndim == 2:
            print("Reshaping rotations array to rank 3.")
            rotations = rotations.reshape(1, *rotations.shape)
        elif rotations.ndim == 3:
            pass
        else:
            raise ValueError("Rotations have to be a rank 2 or 3 array.")
        self._rotations = rotations.astype(self._default_dtype)

    @property
    def target(self):
        """Returns the target NDArray."""
        if type(self._target) == Density:
            return self._target.data
        return self._target

    @property
    def template(self):
        """Returns the reversed template NDArray."""
        if type(self._template) == Density:
            return backend.reverse(self._template.data)
        return backend.reverse(self._template)

    @template.setter
    def template(self, template: NDArray):
        """
        Set the template array.

        Parameters
        ----------
        template : NDArray
            Array to set as the template.
        """
        if type(template) == Density:
            template.data = template.data.astype(self._default_dtype, copy=False)
            self._template = template
            self._templateshape = self._template.shape[::-1]
            return None
        self._template = template.astype(self._default_dtype, copy=False)
        self._templateshape = self._template.shape[::-1]

    @property
    def target_mask(self):
        """Returns the target mask NDArray."""
        if type(self._target_mask) == Density:
            return self._target_mask.data
        return self._target_mask

    @target_mask.setter
    def target_mask(self, mask: NDArray):
        """Sets the target mask."""
        if not np.all(self.target.shape == mask.shape):
            raise ValueError("Target and its mask have to have the same shape.")

        if type(mask) == Density:
            mask.data = mask.data.astype(self._default_dtype, copy=False)
            self._target_mask = mask
            self._targetmaskshape = self._target_mask.shape[::-1]
            return None
        self._target_mask = mask.astype(self._default_dtype, copy=False)
        self._targetmaskshape = self._target_mask.shape

    @property
    def template_mask(self):
        """
        Set the template mask array after reversing it.

        Parameters
        ----------
        template : NDArray
            Array to set as the template.
        """
        if type(self._template_mask) == Density:
            return backend.reverse(self._template_mask.data)
        return backend.reverse(self._template_mask)

    @template_mask.setter
    def template_mask(self, mask: NDArray):
        """Returns the reversed template mask NDArray."""
        if not np.all(self._template.shape == mask.shape):
            raise ValueError("Target and its mask have to have the same shape.")

        if type(mask) == Density:
            mask.data = mask.data.astype(self._default_dtype, copy=False)
            self._template_mask = mask
            self._templatemaskshape = self._template_mask.shape[::-1]
            return None

        self._template_mask = mask.astype(self._default_dtype, copy=False)
        self._templatemaskshape = self._template_mask.shape[::-1]

    def _split_rotations_on_jobs(self, n_jobs: int) -> List[NDArray]:
        """
        Split the rotation matrices into parts based on the number of jobs.

        Parameters
        ----------
        n_jobs : int
            Number of jobs for splitting.

        Returns
        -------
        list of NDArray
            List of split rotation matrices.
        """
        nrot_per_job = self.rotations.shape[0] // n_jobs
        rot_list = []
        for n in range(n_jobs):
            init_rot = n * nrot_per_job
            end_rot = init_rot + nrot_per_job
            if n == n_jobs - 1:
                end_rot = None
            rot_list.append(self.rotations[init_rot:end_rot])
        return rot_list
